"""Module containing the logic for connector adaptor"""

import re

from dgspoc.exceptions import AdaptorAuthenticationError

from dgspoc.constant import ECODE

from dgspoc.utils import MiscOutput


class Adaptor:
    def __init__(self, adaptor, *args, **kwargs):
        self.adaptor = adaptor.strip()
        if self.is_unreal_device_adaptor:
            addr = args[0]
            self.device = UnrealDeviceAdaptor(addr)
        else:
            fmt = '*** Need to implement "{}" adaptor'
            NotImplementedError(fmt.format(adaptor))

    @property
    def is_unreal_device_adaptor(self):
        result = re.match('(?i)unreal-?device$', self.adaptor)
        return bool(result)

    def connect(self):
        result = self.device.connect()
        if self.is_unreal_device_adaptor:
            if self.device.result.startswith('UnrealDeviceConnectionError:'):
                raise AdaptorAuthenticationError(self.device.result)
        return result

    def execute(self, *args, **kwargs):
        result = self.device.execute(*args, **kwargs)
        return result

    def configure(self, *args, **kwargs):
        result = self.device.configure(*args, **kwargs)
        return result

    def disconnect(self, *args, **kwargs):
        result = self.device.disconnect(*args, **kwargs)
        return result

    def reload(self, *args, **kwargs):
        result = self.device.reload(*args, **kwargs)
        return result

    def release(self, *args, **kwargs):
        result = self.device.release(*args, **kwargs)
        return result


class UnrealDeviceAdaptor:
    def __init__(self, address):
        self.address = str(address).strip()
        self.name = self.address
        self.result = ''
        self.exit_code = ECODE.SUCCESS

    @property
    def status(self):
        return True if self.exit_code == ECODE.SUCCESS else False

    def process(self, statement):
        result = MiscOutput.execute_shell_command(statement.strip())
        self.exit_code, self.result = result.exit_code, result.output
        print(self.result)
        return self.status

    def connect(self):
        fmt = 'unreal-device connect --host={}'
        unreal_statement = fmt.format(self.address)
        self.process(unreal_statement)
        return self.status

    def execute(self, *args, **kwargs):
        addr = kwargs.get('name', self.address)
        cmdline = args[0]
        fmt = 'unreal-device execute {} --host={}'
        unreal_statement = fmt.format(cmdline, addr)
        self.process(unreal_statement)
        return self.result

    def configure(self, *args, **kwargs):
        addr = kwargs.get('name', self.address)
        cfg_reference = args[0]
        fmt = 'unreal-device configure {} --host={}'
        unreal_statement = fmt.format(cfg_reference, addr)
        self.process(unreal_statement)
        return self.result

    def disconnect(self, *args, **kwargs):      # noqa
        addr = kwargs.get('name', self.address)
        fmt = 'unreal-device disconnect --host={}'
        unreal_statement = fmt.format(addr)
        self.process(unreal_statement)
        return self.status

    def reload(self, *args, **kwargs):      # noqa
        addr = kwargs.get('name', self.address)
        fmt = 'unreal-device reload --host={}'
        unreal_statement = fmt.format(addr)
        self.process(unreal_statement)
        return self.status

    def release(self, *args, **kwargs):     # noqa
        addr = kwargs.get('name', self.address)
        fmt = 'unreal-device release --host={}'
        unreal_statement = fmt.format(addr)
        self.process(unreal_statement)
        return self.status



"""Module containing the logic for describe-get-system proof of conception entry-points."""

import sys
import argparse

from dgspoc import version
from dgspoc.config import Data

from dgspoc.utils import Printer
from dgspoc.utils import Text

from dgspoc.constant import ECODE
from dgspoc.constant import FEATURE

from dgspoc.usage import validate_usage

from dgspoc.operation import OptionSelector


class ArgumentParser(argparse.ArgumentParser):

    def parse_args(self, *args, **kwargs):
        try:
            options = super().parse_args(*args, **kwargs)
        except BaseException as ex:    # noqa
            if isinstance(ex, SystemExit):
                if ex.code == ECODE.SUCCESS:
                    sys.exit(ECODE.SUCCESS)
                else:
                    self.print_help()
                    sys.exit(ECODE.BAD)
            else:
                Printer.print_message('\n{}\n', Text(ex))
                self.print_help()
                sys.exit(ECODE.BAD)

        if options.help:
            if not options.command:
                self.print_help()
                sys.exit(ECODE.SUCCESS)
            else:
                if options.command in Cli.commands:
                    command = options.command
                    feature = options.operands[0].lower() if options.operands else ''
                    if feature == FEATURE.SCRIPT:
                        name = '{}_script'.format(command)
                    else:
                        name = '{}_{}'.format(command, feature) if feature else command
                    validate_usage(name, ['usage'])
                else:
                    self.print_help()
                    sys.exit(ECODE.BAD)
        return options


class Cli:
    """describe-get-system proof of concept console CLI application."""
    prog = Data.console_cli_name
    prog_fn = Data.console_cli_fullname
    commands = Data.console_supported_commands

    def __init__(self):
        parser = ArgumentParser(
            prog=self.prog,
            usage='%(prog)s [options] command operands',
            description='{} Proof of Concept'.format(self.prog_fn.title()),
            add_help=False
        )

        parser.add_argument(
            '-h', '--help', action='store_true',
            help='show this help message and exit'
        )

        parser.add_argument(
            '-v', '--version', action='version',
            version='%(prog)s v{}'.format(version)
        )

        parser.add_argument(
            '--author', type=str, default='',
            help="author's name"
        ),

        parser.add_argument(
            '--email', type=str, default='',
            help="author's email"
        ),

        parser.add_argument(
            '--company', type=str, default='',
            help="author's company"
        ),

        parser.add_argument(
            '--save-to', type=str, dest='filename', default='',
            help="saving to file"
        ),

        parser.add_argument(
            '--template-id', type=str, dest='tmplid', default='',
            help="template ID"
        ),

        parser.add_argument(
            '--clear', type=str, dest='template_id', default='',
            help="clear template from template-storage"
        ),

        parser.add_argument(
            '--adaptor', type=str, default='',
            help="connector adaptor"
        ),

        parser.add_argument(
            '--action', type=str, default='',
            help="execution action which uses to test template or verification"
        ),

        parser.add_argument(
            '--replaced', action='store_true',
            help='overwrite template ID/file'
        )

        parser.add_argument(
            '--all', action='store_true',
            help='showing all information'
        )

        parser.add_argument(
            '--dependency', action='store_true',
            help='showing package dependency'
        )

        parser.add_argument(
            '--template-storage', action='store_true',
            help='showing template storage information'
        )

        parser.add_argument(
            '--delete', dest='filepath', type=str, default='',
            help="delete file or folder"
        ),

        parser.add_argument(
            '--detail', action='store_true',
            help='generate detail test execution script or report'
        )

        parser.add_argument(
            '--quiet', action='store_true',
            help='silent success or fail info'
        )

        parser.add_argument(
            'command', nargs='?', type=str, default='',
            help='command must be either build, '
                 'info, report, search, test, version, or usage'
        )
        parser.add_argument(
            'operands', nargs='*', type=str,
            help='operands can be template, unittest, '
                 'pytest, robotframework, script, or data such command-line, '
                 'config-lines, or filename'
        )

        self.kwargs = dict()
        self.parser = parser
        self.options = self.parser.parse_args()

    def validate_command(self):
        """Validate argparse `options.command`.

        Returns
        -------
        bool: show ``self.parser.print_help()`` and call ``sys.exit(ECODE.BAD)`` if
        command is neither build, info, report, search,
        test, version, nor usage otherwise, return True
        """
        self.options.command = self.options.command.lower()

        if self.options.command:
            if self.options.command in self.commands:
                return True
            else:
                self.parser.print_help()
                sys.exit(ECODE.BAD)
        return True

    def run(self):
        """Take CLI arguments, parse it, and process."""
        self.validate_command()

        node = OptionSelector(self.options, print_help=self.parser.print_help)
        node.process()


def execute():
    """Execute template console CLI."""
    app = Cli()
    app.run()



"""Module containing robotframework library for describe-get-system proof of concept."""

from dgspoc.core import Dgs


def wait_for(duration):
    """pausing function
    | Parameters:
    |     duration (float): total seconds
    """
    Dgs.wait_for(duration)


def connect_device(host, adaptor='unreal-device', **kwargs):
    """generic function to establish device connection
    | Parameters:
    |     host (str): host name or address
    |     adaptor (str): an adaptor for device connection
    |     kwargs (dict): additional keyword arguments for connecting device connection
    | Returns:
    |     Adaptor: Adaptor connection of device.
    """
    result = Dgs.connect_device(host, adaptor=adaptor, **kwargs)
    return result


def disconnect_device(connection, **kwargs):
    """generic function to disconnect device connection
    | Parameters:
    |     connection (Adaptor): connection of device
    |     kwargs (dict): additional keyword arguments for disconnecting device connection
    | Returns:
    |     bool: True if successfully disconnected device, otherwise, False.
    """
    result = Dgs.disconnect_device(connection, **kwargs)
    return result


def release_device(connection, **kwargs):
    """generic function to release device connection
    | Parameters:
    |     connection (Adaptor): connection of device
    |     kwargs (dict): additional keyword arguments for releasing device connection
    | Returns:
    |     bool: True if successfully released device, otherwise, False.
    """
    result = Dgs.release_device(connection, **kwargs)
    return result


def execute(connection, cmdline, **kwargs):
    """generic function to execute command for device
    | Parameters:
    |     connection (Adaptor): connection of device
    |     cmdline (str): command line
    |     kwargs (dict): additional keyword arguments for command line execution
    | Returns:
    |     str: output of command line
    """
    result = Dgs.execute_cmdline(connection, cmdline, **kwargs)
    return result


def configure(connection, cfg, **kwargs):
    """generic function to configure device
    | Parameters:
    |     connection (Adaptor): connection of device
    |     cfg (str): configuration
    |     kwargs (dict): additional keyword arguments for configuring device
    | Returns:
    |     str: the result of device configuration
    """
    result = Dgs.configure_device(connection, cfg, **kwargs)
    return result


def reload(connection, reload_command, **kwargs):
    """generic function to reload device
    | Parameters:
    |     connection (Adaptor): connection of device
    |     reload_command (str): a reload command
    |     kwargs (dict): additional keyword arguments for reloading device
    | Returns:
    |     str: the result of reloading process
    """
    result = Dgs.reload_device(connection, reload_command, **kwargs)
    return result


def convert_and_filter(text, convertor='', template_ref='', select_statement=''):
    """generic function to convert text data struct and filter per select_statement
    | Parameters:
    |     text (str): output or text data
    |     convertor (str): cvs, json, or template
    |     template_ref (str): template-id or template filename
    |     select_statement (str): a select statement
    | Returns:
    |     list: the list of records
    """
    result = Dgs.convert_and_filter(
        text, convertor=convertor, template_ref=template_ref,
        select_statement=select_statement
    )
    return result



"""Top-level module for describe-get-system proof of concept.

- allow end-user to describe his or her requirement to build test script.
"""

from dgspoc.core import Dgs
from dgspoc.core import sleep
from dgspoc.core import wait_for
from dgspoc.core import connect_device
from dgspoc.core import disconnect_device
from dgspoc.core import release_device
from dgspoc.core import destroy_device
from dgspoc.core import execute
from dgspoc.core import configure
from dgspoc.core import reload
from dgspoc.core import convert_and_filter

from dgspoc.config import version

__version__ = version

__all__ = sorted([
    'Dgs',
    'wait_for',
    'sleep',
    'connect_device',
    'disconnect_device',
    'release_device',
    'destroy_device',
    'execute',
    'configure',
    'reload',
    'convert_and_filter',
    'version',
])



"""Module containing the logic for XML instance as dot object"""

from xml.etree import ElementTree

from dgspoc.utils import DotObject


class DotElement(ElementTree.Element):

    def __init__(self, tag, attrib=None, **extra):
        attrib = attrib or {}

        if 'raw_element' in extra:
            self.raw_element = extra.pop('raw_element')
            extra.update(dict(text=self.raw_element.text,   # noqa
                              tail=self.raw_element.tail))  # noqa
        else:
            self.raw_element = None

        super().__init__(tag, attrib=attrib, **extra)
        self.parent = None

    def __getattribute__(self, item):
        result = super().__getattribute__(item)
        if item == 'attrib':
            if isinstance(result, dict):
                dot_object = DotObject(result)
                return dot_object
            return result
        else:
            return result

    def __repr__(self):
        return "<DotElement %r at %#x>" % (self.tag, id(self))

    def __len__(self):
        return len(self.raw_element) if self.raw_element else len(self)

    def __copy__(self):
        element = super().__copy__()        # noqa
        element.raw_element = self.raw_element
        return element

    def __bool__(self):
        return len(self) != 0

    def __getitem__(self, index):
        if self.raw_element:
            child = self.raw_element[index]
            new_child = to_dot_element(child)
            return new_child
        else:
            child = super().__getitem__(index)
            return child

    def __setitem__(self, index, element):
        if self.raw_element:
            self.raw_element.__setitem__(index, element)
        else:
            super().__setitem__(index, element)

    def __delitem__(self, index):
        if self.raw_element:
            self.raw_element.__delitem__(index)
        else:
            super().__delitem__(index)

    @property
    def has_children(self):
        return len(self.raw_element) if self.raw_element else len(self)

    @property
    def children(self):
        if not self.has_children:
            yield from ()
        cls = self.__class__
        node = self.raw_element if self.raw_element else self
        for child in node:
            if isinstance(child, cls):
                child.parent = self
                yield child
            else:
                kwargs = dict(attrib=child.attrib, raw_element=child)
                new_child = cls(child.tag, **kwargs)
                new_child.parent = self
                yield new_child

    @property
    def raw_children(self):
        if not self.raw_element:
            yield from ()

        for child in self.raw_element:
            yield child

    @property
    def is_leaf(self):
        return not self.has_children

    @property
    def prev_sibling(self):
        try:
            parent = self.parent
            if self.raw_element:
                raw_children = parent.raw_children
                lst = list(raw_children)
                index = lst.index(self.raw_element)
                _prev_sibling = lst[index - 1]
            else:
                lst = list(parent)
                index = lst.index(self)
                _prev_sibling = lst[index - 1]

            return _prev_sibling
        except Exception as ex:     # noqa
            return None

    @property
    def next_sibling(self):
        try:
            parent = self.parent
            if self.raw_element:
                raw_children = parent.raw_children
                lst = list(raw_children)
                index = lst.index(self.raw_element)
                _next_sibling = lst[index + 1]
            else:
                lst = list(parent)
                index = lst.index(self)
                _next_sibling = lst[index + 1]
            return _next_sibling
        except Exception as ex:     # noqa
            return None

    def append(self, subelement):
        if self.raw_element:
            self.raw_element.append(subelement)
        else:
            super().append(subelement)

    def extend(self, elements):
        if self.raw_element:
            self.raw_element.extend(elements)
        else:
            super().extend(elements)

    def insert(self, index, subelement):
        if self.raw_element:
            self.raw_element.insert(index, subelement)
        else:
            super().insert(index, subelement)

    def remove(self, subelement):
        if self.raw_element:
            self.raw_element.remove(subelement)
        else:
            super().remove(subelement)

    def find(self, path, namespaces=None):
        kwargs = dict(namespaces=namespaces)
        if self.raw_element:
            result = self.raw_element.find(path, **kwargs)
        else:
            result = super().find(path, **kwargs)

        for index, item in enumerate(result):
            new_item = to_dot_element(item)
            if new_item != item:
                result[index] = new_item
        return result

    def findtext(self, path, default=None, namespaces=None):
        kwargs = dict(default=default, namespaces=namespaces)
        if self.raw_element:
            result = self.raw_element.findtext(path, **kwargs)
        else:
            result = super().findtext(path, **kwargs)
        return result

    def findall(self, path, namespaces=None):
        kwargs = dict(namespaces=namespaces)
        if self.raw_element:
            result = self.raw_element.findall(path, **kwargs)
        else:
            result = super().findall(path, **kwargs)

        for index, item in enumerate(result):
            new_item = to_dot_element(item)
            if new_item != item:
                result[index] = new_item
        return result

    def iterfind(self, path, namespace=None):
        kwargs = dict(namespace=namespace)
        is_iter = False

        if self.raw_element:
            result = self.raw_element(path, **kwargs)
        else:
            result = super().iterfind(path, **kwargs)

        for node in result:
            is_iter = True
            new_node = to_dot_element(node)
            yield new_node

        if not is_iter:
            yield from ()

    def iter(self, tag=None):
        kwargs = dict(tag=tag)
        is_iter = False

        if self.raw_element:
            result = self.raw_element.iter(**kwargs)
        else:
            result = super().iter(**kwargs)

        for node in result:
            is_iter = True
            new_node = to_dot_element(node)
            yield new_node

        if not is_iter:
            yield from ()

    def clear(self):
        if self.raw_element:
            self.raw_element.clear()
        super().clear()


class DotElementTree(ElementTree.ElementTree):

    def __init__(self, element=None, file=None):
        new_element = to_dot_element(element)
        super().__init__(element=new_element, file=file)

    def getroot(self):
        root = super().getroot()
        kwargs = dict(attrib=root.attrib, raw_element=root)
        new_root = DotElement(root.tag, **kwargs)
        return new_root

    def parse(self, source, parser=None):
        super().parse(source, parser=parser)
        root = self.getroot()
        return root


iselement = ElementTree.iselement


SubElement = ElementTree.SubElement


Comment = ElementTree.Comment


ProcessingInstruction = ElementTree.ProcessingInstruction


PI = ElementTree.PI


register_namespace = ElementTree.register_namespace


def tostring(element, encoding=None, method=None, *,
             xml_declaration=None, default_namespace=None,
             short_empty_elements=True):
    _element = getattr(element, 'raw_element', element)
    kwargs = dict(
        encoding=encoding, method=method,
        xml_declaration=xml_declaration,
        default_namespace=default_namespace,
        short_empty_elements=short_empty_elements
    )
    result = ElementTree.tostring(_element, **kwargs)
    return result


def tostringlist(element, encoding=None, method=None, *,
                 xml_declaration=None, default_namespace=None,
                 short_empty_elements=True):
    _element = getattr(element, 'raw_element', element)
    kwargs = dict(
        encoding=encoding, method=method,
        xml_declaration=xml_declaration,
        default_namespace=default_namespace,
        short_empty_elements=short_empty_elements
    )
    result = ElementTree.tostringlist(_element, **kwargs)
    return result


def dump(elem):
    element = getattr(elem, 'raw_element', elem)
    ElementTree.dump(element)


def parse(source, parser=None):
    tree = DotElementTree()
    tree.parse(source, parser)
    return tree


def to_dot_element(node):
    if isinstance(node, ElementTree.Element):
        kwargs = dict(attrib=node.attrib, raw_element=node)
        new_node = DotElement(node.tag, **kwargs)
        return new_node
    return node



"""Module containing the logic for report of test execution"""

import re
from os import path
from glob import glob

import platform
import robot
import pytest

from datetime import datetime

from collections import OrderedDict

from dgspoc.constant import ECODE
from dgspoc.constant import FWTYPE
from dgspoc.exceptions import ReportError

from dgspoc.utils import DotObject
from dgspoc.utils import Printer

from xml.etree import ElementTree


class DGSReport:
    def __init__(self, *report_files, detail=False):
        if not report_files:
            raise ReportError('CANT generate report without report file.')

        self.report_files = report_files
        self.detail = detail
        self.exit_code = ECODE.PASSED

    def get_report_title(self, framework):      # noqa
        node = DotObject(py_ver=platform.python_version(),
                         rf_ver=robot.__version__,
                         pytest_ver=pytest.__version__,
                         os_name=platform.system())
        if framework == FWTYPE.UNITTEST:
            fmt = 'Unittest Report - Unittest {py_ver} - Python {py_ver} on {os_name}'
        elif framework == FWTYPE.PYTEST:
            fmt = 'Pytest Report - Pytest {pytest_ver} - Python {py_ver} on {os_name}'
        else:
            fmt = 'Robotframework Report - Robot {rf_ver} - Python {py_ver} on {os_name}'

        report_title = fmt.format(**node)
        return report_title

    def generate(self):
        lst = []
        for report_file in self.report_files:
            file_obj = DGSReportFile(report_file)
            if file_obj.is_unittest:
                report_content = self.generate_unittest_report(report_file)
            elif file_obj.is_pytest:
                report_content = self.generate_pytest_report(report_file)
            else:
                report_content = self.generate_robotframework_report(report_file)

            if report_content.strip():
                lst and lst.append('')
                lst.append(report_content)

        all_reports = '\n'.join(lst)
        return all_reports

    def generate_unittest_report(self, file_name):
        tree = ElementTree.parse(file_name)
        root = tree.getroot()
        testsuites = root.findall('testsuite')

        od = OrderedDict()

        total_failed = 0

        for testsuite in testsuites:
            for testcase in testsuite:
                name = testcase.attrib.get('file')
                result = od.get(name, DotObject(total=0, passed=0, failed=0, skipped=0))
                result.name = name
                od[name] = result
                result.total += 1
                if testcase.findall('failure') or testcase.findall('error'):
                    result.failed += 1
                result.passed = result.total - result.failed
            failures = int(testsuite.attrib.get('failures'))
            errors = int(testsuite.attrib.get('errors'))

            if failures or errors:
                total_failed += 1

        if total_failed:
            self.exit_code = ECODE.FAILED

        fmt = ('  - Test case: {name} (Total: {total} / Passed: {passed} / '
               'Failed: {failed} / Skipped: {skipped})')
        lst = []
        for testcase_name, val in od.items():
            lst.append(fmt.format(**val))

        total = len(od)
        total_passed = total - total_failed
        fmt = 'Total Test Cases: %s / Passed: %s / Failed: %s'
        report_summary = fmt % (total, total_passed, total_failed)
        report_title = self.get_report_title(FWTYPE.UNITTEST)
        header = [report_title, '-' * len(report_title), report_summary]
        header_txt = Printer.get(header)

        testcases_txt = '\n'.join(lst)

        if self.detail:
            report = '%s\n%s' % (header_txt, testcases_txt)
        else:
            report = header_txt

        return report

    def generate_pytest_report(self, file_name):
        tree = ElementTree.parse(file_name)
        root = tree.getroot()
        testsuite = root.find('testsuite')

        od = OrderedDict()

        result = None
        total_failed = 0

        for testcase in testsuite:
            name = testcase.attrib.get('classname').rsplit('.', 1)[0]
            result = od.get(name, DotObject(total=0, passed=0, failed=0, skipped=0))
            result.name = name
            od[name] = result
            result.total += 1
            if testcase.findall('failure') or testcase.findall('error'):
                result.failed += 1
            result.passed = result.total - result.failed
        else:
            if result and result.failed:
                total_failed += 1

        if total_failed:
            self.exit_code = ECODE.FAILED

        fmt = ('  - Test case: {name} (Total: {total} / Passed: {passed} / '
               'Failed: {failed} / Skipped: {skipped})')
        lst = []
        for testcase_name, val in od.items():
            lst.append(fmt.format(**val))

        total = len(od)
        total_passed = total - total_failed
        fmt = 'Total Test Cases: %s / Passed: %s / Failed: %s'
        report_summary = fmt % (total, total_passed, total_failed)
        report_title = self.get_report_title(FWTYPE.PYTEST)
        header = [report_title, '-' * len(report_title), report_summary]
        header_txt = Printer.get(header)

        testcases_txt = '\n'.join(lst)

        if self.detail:
            report = '%s\n%s' % (header_txt, testcases_txt)
        else:
            report = header_txt

        return report

    def generate_robotframework_report(self, file_name):
        tree = ElementTree.parse(file_name)
        root = tree.getroot()

        # suite_node = root.find('suite')
        statistics_suite = root.find('statistics/suite')
        errors_node = root.find('errors')

        lst = []

        fmt = ('  - Test case: {name} (Total: {total} / Passed: {passed} / '
               'Failed: {failed} / Skipped: {skipped})')
        total_failed = 0
        result = None
        for index, node in enumerate(statistics_suite):
            if index == 0:
                continue

            result = DotObject(
                name=node.text,
                passed=int(node.attrib.get('pass')),
                failed=int(node.attrib.get('fail')),
                skipped=int(node.attrib.get('skip'))
            )
            result.total = result.passed + result.failed + result.skipped
            lst.append(fmt.format(**result))
        else:
            if result and result.failed:
                total_failed += 1

        if total_failed:
            self.exit_code = ECODE.FAILED

        total = len(statistics_suite) - 1
        total_passed = total - total_failed
        fmt = 'Total Test Cases: %s / Passed: %s / Failed: %s'
        report_summary = fmt % (total, total_passed, total_failed)
        report_title = self.get_report_title(FWTYPE.ROBOTFRAMEWORK)
        header = [report_title, '-' * len(report_title), report_summary]
        header_txt = Printer.get(header)

        testcases_txt = '\n'.join(lst)

        errors = []
        for index, node in enumerate(errors_node):
            error_no = index + 1
            errors.append(Printer.get('Error #%s' % error_no))
            errors.append(node.text)

        errors_txt = '\n'.join(errors)

        if self.detail:
            report = '%s\n%s' % (header_txt, testcases_txt)
            if errors_txt:
                report = '%s\n%s' % (report, errors_txt)
        else:
            report = header_txt

        return report

    @classmethod
    def generate_report(cls, directory=''):
        report_files = DGSReportFile.get_report_files(directory=directory)
        report = cls(*report_files)

        report_content = report.generate()
        return report_content

    
class DGSReportFile:
    def __init__(self, filename):
        self.filename = str(filename).strip()
        self.basename = path.basename(self.filename)

    @property
    def mark_time(self):
        if self.is_report_file:
            if self.is_unittest:
                fmt = 'unittest_report_%Y%b%d_%H%M%S.xml'
            elif self.is_pytest:
                fmt = 'pytest_report_%Y%b%d_%H%M%S.xml'
            else:
                fmt = 'robotframework_output_%Y%b%d_%H%M%S.xml'
            dt = datetime.strptime(self.basename, fmt)
            return dt
        else:
            dt = datetime(1900, 1, 1)
            return dt

    @property
    def is_report_file(self):
        chk = self.is_unittest
        chk |= self.is_pytest
        chk |= self.is_robotframework
        return chk
    
    @property
    def is_unittest(self):
        chk = self.check_report_file(FWTYPE.UNITTEST)
        return chk
    
    @property
    def is_pytest(self):
        chk = self.check_report_file(FWTYPE.PYTEST)
        return chk
    
    @property
    def is_robotframework(self):
        chk = self.check_report_file(FWTYPE.ROBOTFRAMEWORK)
        return chk

    def check_report_file(self, framework):
        other = 'output' if framework == FWTYPE.ROBOTFRAMEWORK else 'report'
        prefix = '%s_%s' % (framework, other)
        fmt = '%s_[0-9]{4}[a-z]{3}[0-9]{2}_[0-9]{6}[.]xml'
        pattern = fmt % prefix
        match = re.match(pattern, self.basename, re.I)
        return bool(match)

    @classmethod
    def get_report_files(cls, directory=''):
        directory = str(directory).strip()

        lookups = ['unittest_report_*_*.xml',
                   'pytest_report_*_*.xml',
                   'robotframework_output_*_*.xml']
        file_names = []
        for lookup in lookups:
            file_path = path.join(directory, lookup)
            file_names += glob(file_path)
        report_files = cls.get_latest_report_files(file_names)

        return report_files

    @classmethod
    def get_latest_report_file(cls, file1, file2):
        file_obj1 = cls(file1)
        file_obj2 = cls(file2)

        if file_obj1.is_report_file and file_obj2.is_report_file:
            if file_obj1.mark_time >= file_obj2.mark_time:
                return file1
            else:
                return file2
        elif file_obj1.is_report_file:
            return file1
        elif file_obj2.is_report_file:
            return file2
        else:
            return ''

    @classmethod
    def get_latest_report_files(cls, file_names):
        if not file_names:
            return file_names

        base_file = file_names[0]
        lst = []
        for file_name in file_names:
            latest_file = cls.get_latest_report_file(base_file, file_name)
            if latest_file and latest_file not in lst:
                lst.append(latest_file)

        if len(lst) > 1:
            latest_dt = cls(lst[0]).mark_time
            for file_name in lst:
                mark_time = cls(file_name).mark_time
                if mark_time >= latest_dt:
                    latest_dt = mark_time

            latest_report_files = []
            for file_name in lst:
                if cls(file_name).mark_time == latest_dt:
                    latest_report_files.append(file_name)
            return latest_report_files
        else:
            return lst



"""Module containing the logic for constant definition"""

import re
from enum import IntFlag


class ICSValue:
    """Treating value as ignore case and ignore space during evaluating
    string equality"""
    def __init__(self, value, equality='', stripped=False):
        self.value = str(value)
        self.equality = equality
        self.stripped = stripped

    def __eq__(self, other):
        value1 = self.value.lower()

        if isinstance(other, self.__class__):
            value2 = other.value.lower()
        else:
            value2 = str(other).lower()

        value1 = re.sub(' +', ' ', value1)
        value2 = re.sub(' +', ' ', value2)

        if self.stripped:
            value1 = value1.strip()
            value2 = value2.strip()

        if self.equality:
            if isinstance(self.equality, (list, tuple)):
                is_equal = True
                for item in self.equality:
                    item = str(item)
                    try:
                        is_equal = bool(re.match(item, value2, re.I))
                    except Exception as ex:     # noqa
                        item = re.sub(' +', ' ', item.lower())
                        is_equal &= item == value2
                return is_equal
            else:
                pattern = str(self.equality)
                try:
                    is_equal = bool(re.match(pattern, value2, re.I))
                except Exception as ex:     # noqa
                    equality = re.sub(' +', ' ', str(self.equality).lower())
                    is_equal = equality == value2
                return is_equal
        else:
            chk = value1.strip() == value2.strip()
        return chk

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return self.value


class ICSStripValue(ICSValue):
    """Treating value as ignore case, ignore space, and strip during evaluating
    string equality"""
    def __init__(self, value, equality=''):
        super().__init__(value, equality=equality, stripped=True)


class ECODE(IntFlag):
    SUCCESS = 0
    BAD = 1
    PASSED = SUCCESS
    FAILED = BAD


class FWTYPE:
    UNITTEST = ICSStripValue('unittest')
    PYTEST = ICSStripValue('pytest')
    ROBOTFRAMEWORK = ICSStripValue('robotframework', equality=r'(rf|robotframework)$')


class CONVTYPE:
    CSV = ICSStripValue('csv')
    JSON = ICSStripValue('json')
    TEMPLATE = ICSStripValue('template')


class COMMAND:
    BUILD = ICSStripValue('build')
    INFO = ICSStripValue('info')
    REPORT = ICSStripValue('report')
    SEARCH = ICSStripValue('search')
    TEST = ICSStripValue('test')
    VERSION = ICSStripValue('version')
    USAGE = ICSStripValue('usage')


class FEATURE:
    TEMPLATE = ICSStripValue('template')
    SCRIPT = ICSStripValue(
        'script', equality=r'(unittest|pytest|robotframework|rf)([_ -]?script)?$'
    )
    BATCH = ICSStripValue('batch')
    UNITTEST = ICSStripValue('unittest', equality=r'unittest([_ -]?script)?$')
    PYTEST = ICSStripValue('pytest', equality=r'pytest([_ -]?script)?$')
    ROBOTFRAMEWORK = ICSStripValue(
        'robotframework', equality=r'(robotframework|rf)([_ -]?script)?$'
    )



"""Module containing the logic for auto generate batch script"""

import re
from datetime import datetime

from dgspoc.utils import File
from dgspoc.utils import DotObject

from dgspoc.constant import FWTYPE


class BatchBuilder:
    def __init__(self, options):
        self.options = options

        self.lookup_fmt = '# %s script is generated by Describe-Get-System'
        self.lookup_pattern = ('# (?P<framework>unitest|pytest|robotframework) '
                               'script is generated by Describe-Get-System')

        self.tbl = DotObject(unittest=[], pytest=[], robotframework=[])

        self.unittest_execution_cmdline = ''
        self.pytest_execution_cmdline = ''
        self.robotframework_execution_cmdline = ''

    @property
    def script(self):
        self.prepare_console_cmdline()

        tbl = DotObject(unittest=[], pytest=[], robotframework=[])
        operands = self.options.operands[1:]
        if not len(operands):
            for case in tbl:
                self.prepare_files('.', case=case, table=tbl)
        else:
            for item in operands:
                item = item.strip()
                if item.lower() in tbl:
                    case = item.lower()
                    self.prepare_files('.', case=case, table=tbl)
                else:
                    pat = (r'(?i)(?P<case>unittest|pytest|robotframework) '
                           r'*= *(?P<directory>\S+.*) *$')
                    match = re.match(pat, item)
                    if match:
                        directory = match.group('directory').strip()
                        case = match.group('case').lower().strip()
                        self.prepare_files(directory, case=case, table=tbl)
                    else:
                        directory = item
                        if File.is_dir(directory):
                            for case in tbl:
                                self.prepare_files(directory, table=tbl, case=case)
                        else:
                            self.prepare_files(directory, table=tbl)

        lst = []
        for fw, files in tbl.items():
            if files:
                if fw == FWTYPE.UNITTEST and self.unittest_execution_cmdline not in lst:
                    lst.append(self.unittest_execution_cmdline)
                elif fw == FWTYPE.PYTEST and self.pytest_execution_cmdline not in lst:
                    lst.append(self.pytest_execution_cmdline)
                else:
                    if self.robotframework_execution_cmdline not in lst:
                        lst.append(self.robotframework_execution_cmdline)

        batch_script = '\n'.join(lst)
        return batch_script

    def prepare_console_cmdline(self):
        dt = datetime.now()
        kwargs = DotObject(prefix='', extension='xml', ref_datetime=dt)
        method = File.get_filepath_timestamp_format1

        # unittest execution
        kwargs.prefix = 'unittest_report'
        fn = method(**kwargs)
        unittest_fmt = 'python -m xmlrunner discover dgs_test_script_files/unittest --output-file=%s'
        self.unittest_execution_cmdline = unittest_fmt % fn

        # pytest execution cmdline
        kwargs.prefix = 'pytest_report'
        fn = method(**kwargs)
        pyttest_fmt = 'python -m pytest --junitxml=%s dgs_test_script_files/pytest'
        self.pytest_execution_cmdline = pyttest_fmt % fn

        # robotframework execution cmdline
        kwargs.prefix = 'robotframework_output'
        output_fn = method(**kwargs)
        kwargs.prefix = 'robotframework_log'
        kwargs.extension = 'html'
        log_fn = method(**kwargs)
        kwargs.prefix = 'robotframework_report'
        report_fn = method(**kwargs)
        robot_fmt = ('python -m robot --output=%s --log=%s --report=%s '
                     'dgs_test_script_files/robotframework')
        self.robotframework_execution_cmdline = robot_fmt % (output_fn, log_fn, report_fn)

    def get_test_script_files(self, directory, case=''):
        filenames = File.get_list_of_filenames(top=directory, pattern='[.](py|robot)$')
        if not filenames:
            result = DotObject(files=[], message=File.message, total=0, case=case)
            return result

        lookup = self.lookup_fmt % case if case else self.lookup_pattern

        lst = []
        for filename in filenames:
            if case:
                chk = File.quicklook(filename, lookup=lookup)
                if chk:
                    lst.append(filename)
            else:
                content = File.get_content(filename)
                match = re.search(lookup, content)
                if match:
                    case = match.group('framework')
                    lst.append(filename)

        result = DotObject(files=lst, message='', total=len(lst), case=case)
        return result

    def prepare_files(self, directory, case='', table=None):
        result = self.get_test_script_files(directory, case=case)
        if result.files:
            case = case or result.case
            dst_dir = 'dgs_test_script_files/%s' % case
            copied_files = File.copy_files(result.files, dst_dir)
            table[case] = list(set(table[case] + copied_files))



"""Module containing the logic for describe-get-system to interpret
user describing problem"""


import re
import operator
# import yaml
from datetime import datetime

from textwrap import indent
from textwrap import dedent
from textwrap import wrap

from dgspoc.utils import DotObject
from dgspoc.utils import Misc
# from dgspoc.utils import File
# from dgspoc.utils import Text

from dgspoc.constant import FWTYPE

# from dgspoc import parser
from dgspoc.parser import ParsedOperation
from dgspoc.parser import CheckStatement

from dgspoc.exceptions import NotImplementedFrameworkError
from dgspoc.exceptions import ComparisonOperatorError
from dgspoc.exceptions import ConnectDeviceStatementError
from dgspoc.exceptions import DisconnectDeviceStatementError
from dgspoc.exceptions import ReleaseDeviceStatementError
from dgspoc.exceptions import WaitForStatementError
from dgspoc.exceptions import PerformerStatementError
from dgspoc.exceptions import VerificationStatementError

from dgspoc.exceptions import ScriptBuilderError


class ScriptInfo(DotObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.devices_vars = dict()
        self._enabled_testing = False

    @property
    def is_testing_enabled(self):
        return self._enabled_testing

    def enable_testing(self):
        self._enabled_testing = True

    def disable_testing(self):
        self._enabled_testing = False

    def get_class_name(self):   # noqa
        return 'TestClass'

    def reset_devices_vars(self):
        self.devices_vars = dict()

    def get_device_var(self, device_name):

        for var_name, var_val in self.devices_vars.items():
            if var_val == device_name:
                return var_name
        else:
            return 'NOT_FOUND_VARIABLE_%s' % device_name


SCRIPTINFO = ScriptInfo()


class Statement:
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):

        self.is_logger = is_logger
        self.data = data
        self.prev = None
        self.next = None
        self.current = None
        self.parent = parent
        self.framework = str(framework).strip()
        self._children = []
        self._name = ''
        self._is_parsed = False

        self._stmt_data = ''
        self._remaining_data = ''

        self._prev_spacers = ''
        self._spacers = ''
        self._level = 0
        self.indentation = indentation

        self.spacer_pattern = r'(?P<spacers> *)[^ ].*'

        self.validate_framework()
        self.prepare()

    def __len__(self):
        return 1 if self.name != '' else 0

    @property
    def is_parsed(self):
        return self._is_parsed

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def children(self):
        return self._children

    @property
    def level(self):
        return self._level

    @property
    def is_unittest(self):
        return self.framework == FWTYPE.UNITTEST

    @property
    def is_pytest(self):
        return self.framework == FWTYPE.PYTEST

    @property
    def is_robotframework(self):
        return self.framework == FWTYPE.ROBOTFRAMEWORK

    @property
    def is_not_robotframework(self):
        return not self.is_robotframework

    @property
    def is_empty(self):
        return self.data.strip() == ''

    @property
    def is_statement(self):
        return self._name != ''

    @property
    def statement_data(self):
        return self._stmt_data

    @property
    def remaining_data(self):
        return self._remaining_data

    @property
    def is_setup_statement(self):
        pattern = r'setup *$'
        is_matched = self.is_matched_statement(pattern)
        return is_matched

    @property
    def is_teardown_statement(self):
        pattern = r'teardown *$'
        is_matched = self.is_matched_statement(pattern)
        return is_matched

    @property
    def is_setup_or_teardown_statement(self):
        return self.is_setup_statement or self.is_teardown_statement

    @property
    def is_parent_setup_or_teardown_statement(self):
        chk = isinstance(self.parent, (SetupStatement, TeardownStatement))
        return chk

    @property
    def is_ancestor_setup_or_teardown_for_unittest(self):
        if not self.is_unittest or not self.parent:
            return False

        node = self.parent

        while isinstance(node, Statement):
            if node.is_setup_or_teardown_statement:
                return True
            node = node.parent
        return False

    @property
    def is_ancestor_base_statement(self):
        if not self.parent:
            return False

        chk_lst = (SetupStatement, SectionStatement, TeardownStatement)

        node = self.parent

        while isinstance(node, Statement):
            if isinstance(node, chk_lst):
                return True
            node = node.parent
        return False

    @property
    def is_ancestor_section_statement(self):
        if not self.parent:
            return False

        node = self.parent

        while isinstance(node, Statement):
            if isinstance(node, SectionStatement):
                return True
            node = node.parent
        return False

    @property
    def is_ancestor_setup_statement(self):
        if not self.parent:
            return False

        if isinstance(self.parent, SetupStatement):
            return True

        node = self.parent

        while isinstance(node, Statement):
            if isinstance(node, SetupStatement):
                return True
            node = node.parent
        return False

    @property
    def is_ancestor_teardown_statement(self):
        if not self.parent:
            return False

        if isinstance(self.parent, TeardownStatement):
            return True

        node = self.parent

        while isinstance(node, Statement):
            if isinstance(node, TeardownStatement):
                return True
            node = node.parent
        return False

    @property
    def is_section_statement(self):
        pattern = r'section'
        is_matched = self.is_matched_statement(pattern)
        return is_matched

    @property
    def is_base_statement(self):
        is_base_stmt = self.is_setup_statement
        is_base_stmt |= self.is_section_statement
        is_base_stmt |= self.is_teardown_statement
        return is_base_stmt

    def is_matched_statement(self, pat, data=None):
        data = data or [self.name, self.statement_data]
        lst = data if Misc.is_list(data) else [data]
        is_matched = any(bool(re.match(pat, str(item), re.I)) for item in lst)
        return is_matched

    def substitute_new_format(self, fmt):
        replacing = '{_replace_}'

        is_ancestor_setup_or_teardown = self.is_ancestor_setup_statement
        is_ancestor_setup_or_teardown |= self.is_ancestor_teardown_statement
        is_ancestor_base_statement = self.is_ancestor_base_statement

        if self.is_unittest or self.is_pytest:
            replaced = ''
            if is_ancestor_base_statement:
                if self.is_unittest and is_ancestor_setup_or_teardown:
                    replaced = 'cls'
                else:
                    replaced = 'self'

            lst = fmt.split(replacing)
            if len(lst) == 1:
                return fmt

            for index, item in enumerate(lst[1:], 1):
                if replaced == '' and item.startswith('.'):
                    lst[index] = item[1:]
            new_fmt = replaced.join(lst)
            return new_fmt
        else:
            if self.parent:
                return fmt
            else:
                lines = fmt.splitlines()
                last_line = lines[-1] if lines else ''
                if re.search('(?i)set +global +variable ', last_line):
                    new_fmt = '\n'.join(lines[:-1])
                    return new_fmt
                else:
                    return fmt

    def prepare(self):
        if self.is_empty:
            self._stmt_data = ''
            self._remaining_data = ''
        else:
            lst = self.data.splitlines()
            for index, line in enumerate(lst):
                line = str(line).rstrip()
                if line.strip():
                    match = re.match(self.spacer_pattern, line)
                    if match:
                        self._spacers = match.group('spacers')
                        length = len(self._spacers)
                        if length == 0:
                            self.set_level(level=0)
                        else:
                            if self.parent:
                                chk_lst = ['setup', 'teardown', 'section']
                                if self.parent.name in chk_lst:
                                    self.set_level(level=1)
                                else:
                                    self.increase_level()
                            else:
                                if self._prev_spacers > self._spacers:
                                    self.increase_level()

                    self._prev_spacers = self._spacers
                    self._stmt_data = line
                    self._remaining_data = '\n'.join(lst[index+1:])

                    if self.is_base_statement:
                        self.set_level(level=0)
                        self._spacers = ''

                    return

    def add_child(self, child):
        if isinstance(child, Statement):
            self._children.append(child)
            if isinstance(child.parent, Statement):
                child.set_level(level=self.level+1)

    def set_level(self, level=0):
        self._level = level

    def increase_level(self):
        self.set_level(level=self.level+1)

    def update_level_from_parent(self):
        if isinstance(self.parent, Statement):
            self.set_level(level=self.parent.level+1)

    def get_next_statement_data(self):
        for line in self.remaining_data.splitlines():
            if line.strip():
                return line
        else:
            return ''

    def has_next_statement(self):
        next_stmt_data = self.get_next_statement_data()
        return next_stmt_data.strip() != ''

    def check_next_statement(self, op):
        op = str(op).strip().lower()
        if op not in ['eq', 'le', 'lt', 'gt', 'ge', 'ne']:
            failure = 'Operator MUST BE eq, ne, le, lt, ge, or gt'
            raise ComparisonOperatorError(failure)

        if not self.has_next_statement():
            return False
        next_stmt_data = self.get_next_statement_data()
        match = re.match(self.spacer_pattern, next_stmt_data)
        spacers = match.group('spacers') if match else ''

        result = getattr(operator, op)(spacers, self._spacers)
        return result

    def is_next_statement_sibling(self):
        result = self.check_next_statement('eq')
        return result

    def is_next_statement_children(self):
        result = self.check_next_statement('gt')
        return result

    def is_next_statement_ancestor(self):
        result = self.check_next_statement('lt')
        return result

    def validate_framework(self):

        if self.framework.strip() == '':
            failure = 'framework MUST be "unittest", "pytest", or "robotframework"'
            raise NotImplementedFrameworkError(failure)

        is_valid_framework = self.is_unittest
        is_valid_framework |= self.is_pytest
        is_valid_framework |= self.is_robotframework

        if not is_valid_framework:
            fmt = ('{!r} framework is not implemented.  It MUST be '
                   '"unittest", "pytest", or "robotframework"')
            raise NotImplementedFrameworkError(fmt.format(self.framework))

    def indent_data(self, data, lvl):
        new_data = indent(data, ' ' * lvl * self.indentation)
        return new_data

    def get_display_method(self, level='info'):
        level = str(level).strip().lower()
        chk_lst = ['debug', 'info', 'warning', 'error', 'fatal', 'critical']
        level = level if level in chk_lst else 'info'
        if self.is_logger:
            if self.is_ancestor_base_statement:
                if self.is_ancestor_section_statement:
                    return 'self.logger.%s' % level
                else:
                    if self.is_unittest:
                        return 'cls.logger.%s' % level
                    else:
                        return 'self.logger.%s' % level
        return 'print'

    def render_display_message(self, message):
        message = getattr(self, 'message', message)
        if not message:
            return ''

        if not self.is_robotframework:
            return message

        message = str(message)
        lst = []
        index = 0
        item = None
        for item in re.finditer(r' +', message):
            lst.append(message[index:item.start()])
            spacers = item.group()
            total = len(spacers)
            lst.append(' ' if total == 1 else '${SPACE * %s}' % total)
            index = item.end()
        else:
            if item:
                lst.append(message[item.end():])
        return ''.join(lst) if lst else message

    def get_display_statement(self, message=''):
        message = self.render_display_message(message)
        method_name = self.get_display_method()
        if self.is_unittest or self.is_pytest:
            stmt = '%s(%r)' % (method_name, message)
        else:   # i.e ROBOTFRAMEWORK
            stmt = 'log   %s' % message

        level = self.parent.level + 1 if self.parent else self.level
        stmt = self.indent_data(stmt, level)
        return stmt

    def get_assert_statement(self, expected_result, assert_only=False):
        is_eresult_number, eresult = Misc.try_to_get_number(expected_result)
        if Misc.is_boolean(eresult):
            eresult = int(eresult)

        if self.is_robotframework:
            fmt1 = 'should be true   True == %s'
            fmt2 = ('${total_count}=   get length ${result}\nshould be '
                    'true   ${result} == %s')
        else:
            if self.is_unittest and self.is_ancestor_section_statement:
                fmt1 = 'self.assertTrue(True == %s)'
                fmt2 = 'total_count = len(result)\nself.assertTrue(total_count == %s)'
            else:
                fmt1 = 'assert True == %s'
                fmt2 = 'total_count = len(result)\nassert total_count == %s'

        fmt = fmt1 if assert_only else fmt2
        eresult = expected_result if assert_only else eresult
        level = self.parent.level + 1 if self.parent else self.level
        stmt = self.indent_data(fmt % eresult, level)
        return stmt

    def try_to_get_base_statement(self):
        if self.is_base_statement:
            tbl = dict(setup=SetupStatement,
                       teardown=TeardownStatement)
            key = self.statement_data.lower().strip()
            cls = tbl.get(key, SectionStatement)
            stmt = cls(self.data, framework=self.framework,
                       indentation=self.indentation,
                       is_logger=self.is_logger)
            return stmt if isinstance(stmt, Statement) else self

        else:
            return self


class DummyStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)
        self.case = ''
        self.message = ''
        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        fmt = 'DUMMY {} - {}'
        expected_result = True if self.case.lower() == 'pass' else False

        message = fmt.format(self.case.upper(), self.message)
        displayed_stmt = self.get_display_statement(message=message)
        assert_stmt = self.get_assert_statement(expected_result, assert_only=True)
        return '{}\n{}'.format(displayed_stmt, assert_stmt)

    def parse(self):
        pattern = ' *dummy[_. -]*(?P<case>pass|fail) *[^a-z0-9]*(?P<message> *.+) *$'
        match = re.match(pattern, self.statement_data, re.I)
        if match:
            self._is_parsed = True
            self.case = match.group('case').lower()
            self.message = match.group('message')
            self.name = 'dummy'
        else:
            self._is_parsed = False


class SetupStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        if self.is_unittest:
            lst.append('@classmethod')
            lst.append('def setUpClass(cls):')
        elif self.is_pytest:
            lst.append('def setup_class(self):')
        else:   # i.e ROBOTFRAMEWORK
            lst.append('setup')

        for child in self.children:
            lst.append(child.snippet)

        level = 0 if self.is_robotframework else 1
        script = self.indent_data('\n'.join(lst), level)
        return script

    def parse(self):
        if self.is_setup_statement:
            self.name = 'setup'
            self._is_parsed = True
            if self.is_next_statement_children():
                node = self.create_child(self)
                self.add_child(node)
                while node and node.is_next_statement_sibling():
                    node = self.create_child(node)
                    self.add_child(node)
                if self.children:
                    last_child = self._children[-1]
                    self._remaining_data = last_child.remaining_data
            if not self.children:
                kwargs = dict(framework=self.framework,
                              indentation=self.indentation,
                              is_logger=self.is_logger)
                data = 'dummy_pass - Dummy Setup'
                dummy_stmt = DummyStatement(data, **kwargs, parent=self)
                self.add_child(dummy_stmt)
        else:
            self._is_parsed = False

    def create_child(self, node):
        kwargs = dict(framework=self.framework,
                      indentation=self.indentation,
                      is_logger=self.is_logger)
        next_line = node.get_next_statement_data()

        if CheckStatement.is_child_connect_device_statement(next_line):
            other = ConnectDeviceStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_dummy_statement(next_line):
            other = DummyStatement(node.remaining_data, **kwargs)
        else:
            return None

        other.prev = node
        # node.next = other
        if isinstance(node, self.__class__):
            other.parent = node
            other.update_level_from_parent()
        else:
            other.parent = node.parent
            other.update_level_from_parent()
        return other


class ConnectDeviceStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.devices_vars = dict()
        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        if not self.has_devices_variables():
            fmt = 'Failed to generate invalid connect device statement - {}'
            failure = fmt.format(self.statement_data)
            raise ConnectDeviceStatementError(failure)

        lst = []
        for var_name, device_name in self.devices_vars.items():
            kwargs = dict(v1=var_name, v2=device_name)
            if self.is_robotframework:
                fmt = ("${%(v1)s}=   connect device   %(v2)s\n"
                       "set global variable   ${%(v1)s}")
                new_fmt = self.substitute_new_format(fmt)
                stmt = new_fmt % kwargs
            else:
                fmt = "{_replace_}.%(v1)s = ta.connect_device(%(v2)r)"
                new_fmt = self.substitute_new_format(fmt)
                stmt = new_fmt % kwargs
            lst.append(stmt)

        level = self.parent.level + 1 if self.parent else self.level
        connect_device_statements = self.indent_data('\n'.join(lst), level)

        return connect_device_statements

    def parse(self):
        pattern = r'(?i) *connect +device +(?P<devices_info>.+) *$'
        match = re.match(pattern, self.statement_data)
        if not match:
            self._is_parsed = False
            return

        devices_info = match.group('devices_info').strip()
        devices_info = devices_info.replace('{', '').replace('}', '')

        pattern = r'(?i)(?P<host>\S+)( +as +(?P<var_name>[a-z]\w*))?$'
        for device_info in devices_info.split(','):
            match = re.match(pattern, device_info.strip())
            if match:
                host, var_name = match.group('host'), match.group('var_name')
                self.reserve_data(host, var_name)
            else:
                fmt = 'Invalid connect device statement - {}'
                failure = fmt.format(self.statement_data)
                raise ConnectDeviceStatementError(failure)

        self.name = 'connect_device'
        self._is_parsed = True

    def reserve_data(self, host, var_name):
        devices_vars = SCRIPTINFO.get('devices_vars', dict())
        SCRIPTINFO.devices_vars = devices_vars

        pattern = r'device[0-9]+$'

        if var_name and str(var_name).strip():
            if var_name not in devices_vars:
                devices_vars[var_name] = host
                self.devices_vars[var_name] = host
            else:
                failure = 'Duplicate device variable - "{}"'.format(var_name)
                raise ConnectDeviceStatementError(failure)
        else:
            var_names = [k for k in devices_vars if re.match(pattern, k)]
            if var_names:
                for v in var_names:
                    if host == devices_vars[v]:
                        self.devices_vars[v] = host
                        return
                new_index = int(var_names[-1].strip('device')) + 1
                key = 'device{}'.format(new_index)
                devices_vars[key] = host
                self.devices_vars[key] = host
            else:
                devices_vars['device1'] = host
                self.devices_vars['device1'] = host

    def has_devices_variables(self):
        return bool(list(self.devices_vars))


class DisconnectStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.vars_lst = []
        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        if not self.vars_lst:
            fmt = 'Failed to generate invalid disconnect device statement - {}'
            failure = fmt.format(self.statement_data)
            raise DisconnectDeviceStatementError(failure)

        lst = []
        for var_name in self.vars_lst:
            kwargs = dict(v1=var_name)
            if self.is_robotframework:
                stmt = "disconnect device   ${%(v1)s}" % kwargs
            else:
                fmt = "ta.disconnect_device({_replace_}.%(v1)s)"
                new_fmt = self.substitute_new_format(fmt)
                stmt = new_fmt % kwargs
            lst.append(stmt)

        level = self.parent.level + 1 if self.parent else self.level
        disconnect_device_statements = self.indent_data('\n'.join(lst), level)

        return disconnect_device_statements

    def parse(self):
        pattern = r'(?i) *disconnect *(device)? +(?P<devices_info>.+) *$'
        match = re.match(pattern, self.statement_data)
        if not match:
            self._is_parsed = False
            return

        devices_info = match.group('devices_info').strip()
        devices_info = devices_info.replace('{', '').replace('}', '')

        pattern = r'(?i)(?P<host>\S+)$'
        for index, device_info in enumerate(devices_info.split(',')):
            match = re.match(pattern, device_info.strip())
            if match:
                host = match.group('host')
                self.reserve_data(host, index)
            else:
                fmt = 'Invalid disconnect device statement - {}'
                failure = fmt.format(self.statement_data)
                raise DisconnectDeviceStatementError(failure)

        self.name = 'disconnect_device'
        self._is_parsed = True

    def reserve_data(self, host, index):
        for var_name, host_name in SCRIPTINFO.devices_vars.items():
            if host == host_name:
                self.vars_lst.append(var_name)
                return

        self.vars_lst.append('device{}'.format(index + 1))


class ReleaseDeviceStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.vars_lst = []
        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        if not self.vars_lst:
            fmt = 'Failed to generate invalid release device statement - {}'
            failure = fmt.format(self.statement_data)
            raise ReleaseDeviceStatementError(failure)

        lst = []
        for var_name in self.vars_lst:
            kwargs = dict(v1=var_name)
            if self.is_robotframework:
                stmt = "release device   ${%(v1)s}" % kwargs
            else:
                fmt = "ta.release_device({_replace_}.%(v1)s)"
                new_fmt = self.substitute_new_format(fmt)
                stmt = new_fmt % kwargs
            lst.append(stmt)

        level = self.parent.level + 1 if self.parent else self.level
        release_device_statements = self.indent_data('\n'.join(lst), level)

        return release_device_statements

    def parse(self):
        pattern = r'(?i) *release +device +(?P<devices_info>.+) *$'
        match = re.match(pattern, self.statement_data)
        if not match:
            self._is_parsed = False
            return

        devices_info = match.group('devices_info').strip()
        devices_info = devices_info.replace('{', '').replace('}', '')

        pattern = r'(?i)(?P<host>\S+)$'
        for index, device_info in enumerate(devices_info.split(',')):
            match = re.match(pattern, device_info.strip())
            if match:
                host = match.group('host')
                self.reserve_data(host, index)
            else:
                fmt = 'Invalid release device statement - {}'
                failure = fmt.format(self.statement_data)
                raise ReleaseDeviceStatementError(failure)

        self.name = 'release_device'
        self._is_parsed = True

    def reserve_data(self, host, index):
        for var_name, host_name in SCRIPTINFO.devices_vars.items():
            if host == host_name:
                self.vars_lst.append(var_name)
                return

        self.vars_lst.append('device{}'.format(index + 1))


class TeardownStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        if self.is_unittest:
            lst.append('@classmethod')
            lst.append('def tearDownClass(cls):')
        elif self.is_pytest:
            lst.append('def teardown_class(self):')
        else:   # i.e ROBOTFRAMEWORK
            lst.append('teardown')

        for child in self.children:
            lst.append(child.snippet)

        level = 0 if self.is_robotframework else 1
        script = self.indent_data('\n'.join(lst), level)
        return script

    def parse(self):
        if self.is_teardown_statement:
            self.name = self.statement_data.strip().lower()
            self._is_parsed = True
            if self.is_next_statement_children():
                node = self.create_child(self)
                self.add_child(node)
                while node and node.is_next_statement_sibling():
                    node = self.create_child(node)
                    self.add_child(node)
                if self.children:
                    last_child = self._children[-1]
                    self._remaining_data = last_child.remaining_data
            if not self.children:
                kwargs = dict(framework=self.framework,
                              indentation=self.indentation,
                              is_logger=self.is_logger)
                data = 'dummy_pass - Dummy %s' % self.name.title()
                dummy_stmt = DummyStatement(data, **kwargs, parent=self)
                self.add_child(dummy_stmt)
        else:
            self._is_parsed = False

    def create_child(self, node):
        kwargs = dict(framework=self.framework,
                      indentation=self.indentation,
                      is_logger=self.is_logger)
        next_line = node.get_next_statement_data()

        if CheckStatement.is_child_disconnect_device_statement(next_line):
            other = DisconnectStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_release_device_statement(next_line):
            other = ReleaseDeviceStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_dummy_statement(next_line):
            other = DummyStatement(node.remaining_data, **kwargs)
        else:
            return None

        other.prev = node
        # node.next = other
        if isinstance(node, self.__class__):
            other.parent = node
            other.update_level_from_parent()
        else:
            other.parent = node.parent
            other.update_level_from_parent()
        return other


class SectionStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.description = ''
        self._method_name = ''
        self.parse()

    @property
    def method_name(self):
        return self._method_name

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        fmt = '%s' if self.is_robotframework else 'def %s(self):'
        lst.append(fmt % self.method_name)

        if self.description and self.description.strip():
            if self.is_robotframework:
                lst1 = wrap(self.description, width=56)
                for index, item in enumerate(lst1):
                    prefix = '[Documentation]' if index == 0 else '...'
                    lst1[index] = '{:18} {}'.format(prefix, item)

                method_doc = '\n'.join(lst1)
                lst.append(self.indent_data(method_doc, 1))
                pass
            else:
                method_doc = '"""%s"""' % self.description
                method_doc = '\n'.join(wrap(method_doc, width=70))
                lst.append(self.indent_data(method_doc, 1))

        for child in self.children:
            lst.append(child.snippet)

        level = 0 if self.is_robotframework else 1
        script = self.indent_data('\n'.join(lst), level)
        return script

    def parse(self):
        if not self.is_section_statement:
            self._is_parsed = False
            return

        pattern = r'(?i) *section([^a-z0-9]+)?(?P<description>\w+.+)?'
        match = re.match(pattern, self.statement_data)
        description = match.group('description') if match else 'test_default'
        self.parse_description(description)

        self.name = 'section'
        self._is_parsed = True
        if self.is_next_statement_children():
            node = self.create_child(self)
            self.add_child(node)
            while node and node.is_next_statement_sibling():
                node = self.create_child(node)
                self.add_child(node)
            if self.children:
                last_child = self._children[-1]
                self._remaining_data = last_child.remaining_data
        if not self.children:
            kwargs = dict(framework=self.framework,
                          indentation=self.indentation,
                          is_logger=self.is_logger)
            data = 'dummy_pass - Dummy for section'
            dummy_stmt = DummyStatement(data, **kwargs, parent=self)
            self.add_child(dummy_stmt)

    def parse_description(self, description):
        if not description or description == 'test_default':
            self.description = 'test default'
            self._method_name = 'test default' if self.is_robotframework else 'test_default'
        else:
            description = ' '.join(str(description).splitlines()).strip()
            pattern = r'(?i)(?P<desc>.+?)( +as +(?P<ref>[a-z]\w*( +\w+)?))?$'
            match = re.match(pattern, description)
            desc, ref = match.group('desc'), match.group('ref')
            ref = ref or desc
            ref = re.sub('(?i)[^a-z0-9]+', '_', ref).strip('_')

            self.description = desc

            if not ref.lower().startswith('test'):
                ref = 'test_%s' % ref

            if self.is_robotframework:
                self._method_name = ref.replace('_', ' ')
            else:
                if len(ref) > 60:
                    ref = wrap(ref.replace('_', ' '), width=60)[0].replace(' ', '_')
                self._method_name = ref

    def create_child(self, node):
        kwargs = dict(framework=self.framework,
                      indentation=self.indentation,
                      is_logger=self.is_logger)
        next_line = node.get_next_statement_data()

        if CheckStatement.is_child_verification_statement(next_line):
            other = VerificationStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_performer_statement(next_line):
            other = PerformerStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_connect_device_statement(next_line):
            other = ConnectDeviceStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_disconnect_device_statement(next_line):
            other = DisconnectStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_release_device_statement(next_line):
            other = ReleaseDeviceStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_pausing_statement(next_line):
            other = WaitForStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_iterative_statement(next_line):
            other = LoopStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_dummy_statement(next_line):
            other = DummyStatement(node.remaining_data, **kwargs)
        else:
            return None

        other.prev = node
        # node.next = other
        if isinstance(node, self.__class__):
            other.parent = node
            other.update_level_from_parent()
        else:
            other.parent = node.parent
            other.update_level_from_parent()
        return other


class LoopStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.ntimes = 0
        self.parse()

    @property
    def is_regular_iterative(self):
        result = CheckStatement.is_regular_iterative_statement(self.statement_data)
        return result

    @property
    def is_util_iterative(self):
        result = CheckStatement.is_until_iterative_statement(self.statement_data)
        return result

    @property
    def is_to_last_iterative(self):
        result = CheckStatement.is_to_last_iterative_statement(self.statement_data)
        return result

    @property
    def is_iterative_statement(self):
        result = self.is_regular_iterative
        result |= self.is_util_iterative
        result |= self.is_to_last_iterative
        return result

    @property
    def snippet(self):
        if self.is_regular_iterative:
            stmt = self.regular_iterative_snippet
            return stmt
        elif self.is_util_iterative:
            stmt = self.until_iterative_snippet
            return stmt
        else:
            stmt = self.to_last_iterative_snippet
            return stmt

    @property
    def regular_iterative_snippet(self):
        if not self.is_parsed:
            return ''
        if self.ntimes <= 0:
            return ''

        lst = []
        if self.is_robotframework:
            lst.append('${ntimes}=   set variable   %s' % self.ntimes)
            lst.append('@{indexes}=   evaluate   range(1, ${ntimes} + 1)')
            lst.append('FOR   ${index}   IN   @{indexes}')
            for child in self.children:
                if child.name == 'verification':
                    msg = 'Failed at iteration ${index}/${ntimes}'
                    child_snippet = child.render_assertion_message(msg)
                    lst.append(child_snippet)
                else:
                    lst.append(child.snippet)
            lst.append('END')
        else:
            lst.append('ntimes = %s' % self.ntimes)
            lst.append('indexes = range(1, times + 1)')
            lst.append('for index in indexes:')
            for child in self.children:
                if child.name == 'verification':
                    msg = "'Failed at iteration {}/{}'.format(index, ntimes)"
                    child_snippet = child.render_assertion_message(msg)
                    lst.append(child_snippet)
                else:
                    lst.append(child.snippet)

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    @property
    def until_iterative_snippet(self):
        if not self.is_parsed:
            return ''
        if self.ntimes <= 0:
            return ''

        lst = []
        if self.is_robotframework:
            lst.append('${ntimes}=   set variable   %s' % self.ntimes)
            lst.append('@{indexes}=   evaluate   range(1, ${ntimes} + 1)')
            lst.append('${is_passed}=   set variable   ${True}')
            lst.append('FOR   ${index}   IN   @{indexes}')
            for child in self.children:
                if child.name == 'verification':
                    msg = 'Failed at iteration ${index}/${ntimes}'
                    child_snippet = child.render_assertion_message(msg)
                    lst.append(child_snippet)
                else:
                    lst.append(child.snippet)

            lst.append(self.indent_data('exit for loop if   ${is_passed}', 1))
            lst.append('END')
        else:
            lst.append('ntimes = %s' % self.ntimes)
            lst.append('indexes = range(1, times + 1)')
            lst.append('is_passed = True')
            lst.append('for index in indexes:')
            for child in self.children:
                if child.name == 'verification':
                    msg = "'Failed at iteration {}/{}'.format(index, ntimes)"
                    child_snippet = child.render_assertion_message(msg)
                    lst.append(child_snippet)
                else:
                    lst.append(child.snippet)

            lst.append(self.indent_data('if is_passed:', 1))
            lst.append(self.indent_data('break', 2))

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    @property
    def to_last_iterative_snippet(self):
        if not self.is_parsed:
            return ''
        if self.ntimes <= 0:
            return ''

        lst = []
        if self.is_robotframework:
            lst.append('${ntimes}=   set variable   %s' % self.ntimes)
            lst.append('@{indexes}=   evaluate   range(1, ${ntimes} + 1)')
            lst.append('${is_passed}=   set variable   ${True}')
            lst.append('FOR   ${index}   IN   @{indexes}')
            for child in self.children:
                if child.name == 'verification':
                    addition = '${is_passed}=   evaluate   ${is_passed} and ${check}'
                    child_snippet = child.convert_assertion_to_check(addition=addition)
                    lst.append(child_snippet)
                else:
                    lst.append(child.snippet)
            lst.append(('run keyword if   ${is_passed} == False   Log   '
                        'failed verification(s) at iteration '
                        '${index}/${ntimes}   WARN'))
            lst.append('END')
            lst.append('should be true   ${is_passed}')
        else:
            lst.append('ntimes = %s' % self.ntimes)
            lst.append('indexes = range(1, times + 1)')
            lst.append('is_passed = True')
            lst.append('for index in indexes:')
            for child in self.children:
                if child.name == 'verification':
                    addition = 'is_passed = is_passed and check'
                    child_snippet = child.convert_assertion_to_check(addition=addition)
                    lst.append(child_snippet)
                else:
                    lst.append(child.snippet)

            lst.append(self.indent_data('if not is_passed:', 1))
            warned_msg = "'Warning: failed verification(s) at iteration {}/{}'.format(index, ntimes)"
            lst.append(self.indent_data('print(%s)' % warned_msg, 2))

            if self.is_unittest and self.is_ancestor_section_statement:
                lst.append('self.assertTrue(is_passed)')
            else:
                lst.append('assert is_passed')

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    def parse(self):
        if not self.is_iterative_statement:
            self._is_parsed = False
            return

        pattern = r'(?i) *loop +(?P<ntimes>[0-9]+) +'
        match = re.match(pattern, self.statement_data)
        self.ntimes = int(match.group('ntimes'))

        if self.is_next_statement_children():
            node = self.create_child(self)
            self.add_child(node)
            while node and node.is_next_statement_sibling():
                node = self.create_child(node)
                self.add_child(node)
            if self.children:
                last_child = self._children[-1]
                self._remaining_data = last_child.remaining_data
        if not self.children:
            kwargs = dict(framework=self.framework,
                          indentation=self.indentation,
                          is_logger=self.is_logger)
            data = 'dummy_pass - Dummy iterative statement'
            dummy_stmt = DummyStatement(data, **kwargs, parent=self)
            self.add_child(dummy_stmt)

        self.name = 'loop'
        self._is_parsed = True
        self.update_level_from_parent()

    def create_child(self, node):
        kwargs = dict(framework=self.framework,
                      indentation=self.indentation,
                      is_logger=self.is_logger)
        next_line = node.get_next_statement_data()

        if CheckStatement.is_child_verification_statement(next_line):
            other = VerificationStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_performer_statement(next_line):
            other = PerformerStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_connect_device_statement(next_line):
            other = ConnectDeviceStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_disconnect_device_statement(next_line):
            other = DisconnectStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_release_device_statement(next_line):
            other = ReleaseDeviceStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_pausing_statement(next_line):
            other = WaitForStatement(node.remaining_data, **kwargs)
        elif CheckStatement.is_child_dummy_statement(next_line):
            other = DummyStatement(node.remaining_data, **kwargs)
        else:
            return None

        other.prev = node
        # node.next = other
        if isinstance(node, self.__class__):
            other.parent = node
            other.update_level_from_parent()
        else:
            other.parent = node.parent
            other.update_level_from_parent()
        return other


class PerformerStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.result = None
        self.parse()

    @property
    def snippet(self):
        if self.name == 'execution':
            stmt = self.execution_snippet
            return stmt
        elif self.name == 'configuration':
            stmt = self.configuration_snippet
            return stmt
        elif self.name == 'reload':
            stmt = self.reload_snippet
            return stmt

    @property
    def execution_snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        result = self.result

        if self.is_robotframework:
            for device_name in result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)
                if result.has_select_statement:
                    fmt = '${output}=   execute   ${%s}   cmdline=%s'
                    lst.append(fmt % (var_name, result.operation_ref))
                    if result.is_template:
                        fmt = ('convert_and_filter\n'
                               '...   ${output}   convertor=%s   template_ref=%s\n'
                               '...   select_statement=%s')
                        stmt = fmt % (result.convertor, result.convertor_arg,
                                      result.select_statement)
                        lst.append(stmt)
                    else:
                        fmt = ('convert_and_filter\n'
                               '...   ${output}   convertor=%s\n'
                               '...   select_statement=%s')
                        stmt = fmt % (result.convertor, result.select_statement)
                        lst.append(stmt)
                else:
                    fmt = 'execute   ${%s}   cmdline=%s'
                    lst.append(fmt % (var_name, self.result.operation_ref))
        else:
            for device_name in self.result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)

                if result.has_select_statement:
                    fmt = 'output = ta.execute({_replace_}.%s, cmdline=%r)'
                    new_fmt = self.substitute_new_format(fmt)
                    lst.append(new_fmt % (var_name, result.operation_ref))
                    if result.is_template:
                        fmt = ('ta.convert_and_filter(\n'
                               '    output, convertor=%r, template_ref=%r,\n'
                               '    select_statement=%r\n'
                               ')')
                        fmt = fmt.replace('    ', ' ' * self.indentation)
                        stmt = fmt % (result.convertor, result.convertor_arg,
                                      result.select_statement)
                        lst.append(stmt)
                    else:
                        fmt = ('ta.convert_and_filter(\n'
                               '    output, convertor=%r,\n'
                               '    select_statement=%r\n'
                               ')')
                        fmt = fmt.replace('    ', ' ' * self.indentation)
                        stmt = fmt % (result.convertor, result.select_statement)
                        lst.append(stmt)
                else:
                    fmt = 'ta.execute({_replace_}.%s, cmdline=%r)'
                    new_fmt = self.substitute_new_format(fmt)
                    lst.append(new_fmt % (var_name, result.operation_ref))

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    @property
    def configuration_snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        result = self.result

        if self.is_robotframework:
            for device_name in result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)
                fmt = 'configure   ${%s}   cfg=%s'
                lst.append(fmt % (var_name, self.result.operation_ref))
        else:
            for device_name in self.result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)

                fmt = 'ta.configure({_replace_}.%s, cfg=%r)'
                new_fmt = self.substitute_new_format(fmt)
                lst.append(new_fmt % (var_name, result.operation_ref))

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    @property
    def reload_snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        result = self.result

        if self.is_robotframework:
            for device_name in result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)
                fmt = 'reload   ${%s}   cmdline=%s'
                lst.append(fmt % (var_name, self.result.operation_ref))
        else:
            for device_name in self.result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)

                fmt = 'ta.reload({_replace_}.%s, cmdline=%r)'
                new_fmt = self.substitute_new_format(fmt)
                lst.append(new_fmt % (var_name, result.operation_ref))

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    def parse(self):
        if not CheckStatement.is_performer_statement(self.statement_data):
            self._is_parsed = False
            return

        result = ParsedOperation(self.statement_data)
        self.result = result
        self._is_parsed = result.is_parsed
        self.name = self.result.name
        self.update_level_from_parent()

        if result.error:
            raise PerformerStatementError(result.error)


class VerificationStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)

        self.result = None
        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        lst = []

        result = self.result

        if self.is_robotframework:
            for device_name in result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)

                fmt = '${output}=   execute   ${%s}   cmdline=%s'
                lst.append(fmt % (var_name, result.operation_ref))
                if result.is_template:
                    fmt = ('${result}=   convert_and_filter\n'
                           '...   ${output}   convertor=%s   template_ref=%s\n'
                           '...   select_statement=%s')
                    stmt = fmt % (result.convertor, result.convertor_arg,
                                  result.select_statement)
                    lst.append(stmt)
                else:
                    fmt = ('${result}=   convert_and_filter\n'
                           '...   ${output}   convertor=%s\n'
                           '...   select_statement=%s')
                    stmt = fmt % (result.convertor, result.select_statement)
                    lst.append(stmt)

                lst.append('${total_count}=  get length   ${result}')
                fmt = 'should be true   ${total_count} %s %s'
                lst.append(fmt % (result.condition_symbol, result.expected_condition))

        else:
            for device_name in self.result.devices_names:
                var_name = SCRIPTINFO.get_device_var(device_name)

                fmt = 'output = ta.execute({_replace_}.%s, cmdline=%r)'
                new_fmt = self.substitute_new_format(fmt)
                lst.append(new_fmt % (var_name, result.operation_ref))
                if result.is_template:
                    fmt = ('result = ta.convert_and_filter(\n'
                           '    output, convertor=%r, template_ref=%r,\n'
                           '    select_statement=%r\n'
                           ')')
                    fmt = fmt.replace('    ', ' ' * self.indentation)
                    stmt = fmt % (result.convertor, result.convertor_arg,
                                  result.select_statement)
                    lst.append(stmt)
                else:
                    fmt = ('result = ta.convert_and_filter(\n'
                           '    output, convertor=%r,\n'
                           '    select_statement=%r\n'
                           ')')
                    fmt = fmt.replace('    ', ' ' * self.indentation)
                    stmt = fmt % (result.convertor, result.select_statement)
                    lst.append(stmt)

                lst.append('total_count = len(result)')
                if self.is_unittest and self.is_ancestor_section_statement:
                    fmt = 'self.assertTrue(total_count %s %s)'
                    lst.append(fmt % (result.condition_symbol, result.expected_condition))
                else:
                    fmt = 'assert total_count %s %s'
                    lst.append(fmt % (result.condition_symbol, result.expected_condition))

        stmt = self.indent_data('\n'.join(lst), self.level)
        return stmt

    def render_assertion_message(self, msg):
        if not self.snippet:
            return ''

        lines = dedent(self.snippet).splitlines()
        pattern = '(?i)(?P<case1>should be true)|(?P<case2>self[.])?assert(True)?'

        lst = []
        for line in lines:
            match = re.match(pattern, line)
            if match:
                txt = line
                if match.group('case1'):
                    fmt = '%s   %s'
                elif match.group('case2'):
                    fmt = '%s, msg=%s)'
                    txt = line[-1]
                else:
                    fmt = '%s, %s'
                new_line = fmt % (txt, msg)
                lst.append(new_line)
            else:
                lst.append(line)

        new_snippet = self.indent_data('\n'.join(lst), self.level)
        return new_snippet

    def convert_assertion_to_check(self, addition=''):
        if not self.snippet:
            return ''

        lines = dedent(self.snippet).splitlines()
        pattern = ('(?i)(?P<case>(should be true)|(self[.])?assert(True)?) *'
                   '(?P<val>[^ ].*[^ ]?) *$')

        lst = []
        for index, line in enumerate(lines):
            match = re.match(pattern, line)
            if match:
                val = match.group('val').strip()
                if self.is_robotframework:
                    lst.append('${check}=   evaluate   %s' % val)
                else:
                    val = val.lstrip('(').rstrip(')')
                    lst.append('check = %s' % val)
                addition and lst.append(addition)
            else:
                lst.append(line)

        new_snippet = self.indent_data('\n'.join(lst), self.level)
        return new_snippet

    def parse(self):
        if not CheckStatement.is_execute_cmdline(self.statement_data):
            self._is_parsed = False
            return

        result = ParsedOperation(self.statement_data)
        self.result = result
        self._is_parsed = result.is_parsed
        self.name = 'verification'
        self.update_level_from_parent()

        if result.error:
            raise PerformerStatementError(result.error)

        if not self.result.is_verification:
            fmt = 'Invalid verification statement format\n %s'
            raise VerificationStatementError(fmt % self.statement_data)


class WaitForStatement(Statement):
    def __init__(self, data, parent=None, framework='',
                 indentation=4, is_logger=False):
        super().__init__(data, parent=parent, framework=framework,
                         indentation=indentation, is_logger=is_logger)
        self.total_seconds = 0
        self.parse()

    @property
    def snippet(self):
        if not self.is_parsed:
            return ''

        fmt = 'wait for   %s' if self.is_robotframework else 'ta.wait_for(%s)'
        stmt = self.indent_data(fmt % self.total_seconds, self.level)
        return stmt

    def parse(self):
        pattern = r'(?i) *((wait +for)|sleep) +(?P<capture_data>[0-9].+) *$'
        match = re.match(pattern, self.statement_data)
        if not match:
            self._is_parsed = False
            return

        capture_data = match.group('capture_data').strip()

        pattern = ('(?P<val>([0-9]*[.])?[0-9]+) *'
                   '(?P<unit>h((ou)?rs?)?|m(in(utes?)?)?|'
                   's(ec(onds?)?)?|d(ays?)?)?')
        match = re.match(pattern, capture_data, re. I)
        if not match:
            failure = 'Invalid wait for statement format'
            raise WaitForStatementError(failure)

        result = DotObject(match.groupdict())
        tbl = dict(s=1, m=60, h=60 * 60, d=60 * 60 * 24)
        multiplier = tbl.get(str(result.unit).lower()[:1], 1)
        seconds = float(result.val) * multiplier
        self.total_seconds = int(seconds) if int(seconds) == seconds else seconds
        self._is_parsed = True
        self.name = 'wait_for'
        self.update_level_from_parent()


class ScriptBuilder:
    def __init__(self, data, framework='', indentation=4, is_logger=False,
                 username='', email='', company='',):
        self.data = data
        self.framework = str(framework).strip()
        self.indentation = indentation
        self.is_logger = is_logger
        self.username = str(username).strip()
        self.email = str(email).strip()
        self.company = str(company).strip() or self.username

        self.setup_statement = None
        self.teardown_statement = None
        self.section_statements = []
        self.build()

    @property
    def testscript(self):
        if self.setup_statement and self.teardown_statement:
            lst = [self.import_library_code,
                   '',
                   self.setup_teardown_code,
                   '',
                   self.sections_code]
            script = '\n'.join(lst).strip()

            if self.framework == FWTYPE.UNITTEST:
                spacer = ' ' * self.indentation
                addition = '\n'.join([
                    "if __name__ == '__main__':",
                    '%s%s' % (spacer, 'unittest.main('),
                    '%s%s' % (spacer * 2, 'testRunner=XMLTestRunner(output="report"),'),
                    '%s%s' % (spacer * 2, 'failfast=False, buffer=False, catchbreak=False'),
                    '%s%s' % (spacer, ')')
                ])
                script = '%s\n\n%s' % (script, addition)

            return script + '\n'
        else:
            if self.setup_statement:
                failure = 'CANT build script without Setup statement'
            elif self.teardown_statement:
                failure = 'CANT build test script without Teardown statement'
            else:
                failure = 'CANT build script without Setup and Teardown statements'
            raise ScriptBuilderError(failure)

    @property
    def get_logger_function(self):
        func_text = dedent('''
            def get_logger(name='TATestScript'):
                """This function only creates logger instance with
                basic logging configuration.
                ==================================================
                PLEASE UPDATE your get_logger function.
                ==================================================
                """
                import logging
                logging.basicConfig(
                    level=logging.INFO,
                    format="%(asctime)s [%(levelname)s] %(message)s",
                    handlers=[
                        logging.FileHandler('%s.log' % name.lower()),
                        logging.StreamHandler()
                    ]
                )
                logger = logging.getLogger(name)
                return logger
        ''').strip()

        func_text = func_text.replace('    ', ' ' * self.indentation)
        return func_text

    @property
    def intro_code(self):
        fmt = '# {} script is generated by Describe-Get-System Proof of Concept'
        user_fmt = '# Created by  : {0.username}'
        email_fmt = '# Email       : {0.email}'
        company_fmt = '# Company     : {0.company}'
        datetime_str = '# Created date: {:%Y-%m-%d}'.format(datetime.now())
        lst = [fmt.format(self.framework.lower())]
        self.username and lst.append(user_fmt.format(self))
        self.email and lst.append(email_fmt.format(self))
        self.company and lst.append(company_fmt.format(self))
        not SCRIPTINFO.is_testing_enabled and lst.append(datetime_str)

        intro = '\n'.join(['#' * 80] + lst + ['#' * 80])
        return intro

    @property
    def import_library_code(self):
        lst = [self.intro_code, '']

        is_unittest = self.framework == FWTYPE.UNITTEST

        if self.framework == FWTYPE.ROBOTFRAMEWORK:
            lst += [
                '*** Settings ***',
                'Library          BuiltIn',
                'Library          Collections',
                'Library          dgspoc.robotframeworklib',
                'Suite Setup      {}'.format(self.setup_statement.name),
                'Suite Teardown   {}'.format(self.teardown_statement.name),
            ]
        else:
            cls_name = SCRIPTINFO.get_class_name()
            inherit = '(unittest.TestCase)' if is_unittest else ''
            lst.append('import unittest' if is_unittest else '# import pytest')
            lst.append('import dgspoc as ta')
            is_unittest and lst.append('from xmlrunner import XMLTestRunner')
            lst.append('\n')
            if self.is_logger:
                lst.append(self.get_logger_function)
                lst.append('\n')
            lst.append('class %s%s:' % (cls_name, inherit))
            if self.is_logger:
                lst.append(' ' * self.indentation + 'logger = get_logger()')

        import_lib_txt = '\n'.join(lst)
        return import_lib_txt

    @property
    def setup_teardown_code(self):

        lst = [self.setup_statement.snippet,
               '',
               self.teardown_statement.snippet]
        if self.framework == FWTYPE.ROBOTFRAMEWORK:
            lst.insert(0, '*** Keywords ***')

        setup_teardown_txt = '\n'.join(lst)
        return setup_teardown_txt

    @property
    def sections_code(self):
        is_robotframework = self.framework == FWTYPE.ROBOTFRAMEWORK
        lst = []
        fmt = 'test %03i ' if is_robotframework else 'def test_%03i_'
        replacing = 'test ' if is_robotframework else 'def test_'

        for index, stmt in enumerate(self.section_statements, 1):
            lst.append('')
            if stmt.snippet:
                lst.append(stmt.snippet.replace(replacing, fmt % index, 1))

        if lst:
            if is_robotframework:
                lst[0] = '*** Test Cases ***'
            not lst[0] and lst.pop(0)

        sections_code_txt = '\n'.join(lst)
        return sections_code_txt

    def build(self):
        data = self.data
        count = 2000

        while data.strip() and count > 0:
            stmt = Statement(data, framework=self.framework,
                             indentation=self.indentation,
                             is_logger=self.is_logger)
            stmt = stmt.try_to_get_base_statement()
            self.add_statement(stmt)
            data = stmt.remaining_data
            count -= 1

    def add_statement(self, stmt):
        if stmt.is_setup_statement:
            if not self.setup_statement:
                self.setup_statement = stmt
            else:
                self.warn_duplicate_statement(stmt)
        elif stmt.is_teardown_statement:
            if not self.teardown_statement:
                self.teardown_statement = stmt
            else:
                self.warn_duplicate_statement(stmt)
        elif stmt.is_section_statement:
            if self.is_uniq_section_statement(stmt):
                self.section_statements.append(stmt)
            else:
                self.warn_duplicate_statement(stmt)
        else:
            self.warn_not_implement_statement(stmt)

    def is_uniq_section_statement(self, stmt):
        if not self.section_statements:
            return True

        chk = stmt.snippet
        is_duplicate = any(chk == k.snippet for k in self.section_statements)
        return not is_duplicate

    def warn_duplicate_statement(self, stmt):
        fmt = 'IncompleteTask - Need to implement warn_duplicate_statement\n{}'
        raise NotImplementedError(fmt.format(stmt.statement_data))

    def warn_not_implement_statement(self, stmt):
        fmt = 'IncompleteTask - Need to implement warn_not_implement_statement\n{}'
        raise NotImplementedError(fmt.format(stmt.statement_data))



"""Packaging dgspoc."""

from setuptools import setup, find_packages


setup(
    name='dgspoc',
    version='0.3.10',
    license='BSD-3-Clause',
    license_files=['LICENSE'],
    description='The proof of concept for Describe-Get-System.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Tuyen Mathew Duong',
    author_email='tuyen@geekstrident.com',
    maintainer='Tuyen Mathew Duong',
    maintainer_email='tuyen@geekstrident.com',
    install_requires=[
        'compare_versions',
        'python-dateutil',
        'textfsm',
        'pyyaml',
        'pytest',
        'robotframework',
        'dlapp',
        'regexapp',
        'templateapp',
        'gtunrealdevice',
        'unittest-xml-reporting'
    ],
    url='https://github.com/Geeks-Trident-LLC/dgspoc',
    packages=find_packages(
        exclude=(
            'tests*', 'testing*', 'examples*',
            'build*', 'dist*', 'docs*', 'venv*'
        )
    ),
    include_package_data=True,
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'dgs = dgspoc.main:execute',
            'dgspoc = dgspoc.main:execute',
        ]
    },
    classifiers=[
        # development status
        'Development Status :: 2 - Pre-Alpha',
        # natural language
        'Natural Language :: English',
        # intended audience
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Manufacturing',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        # operating system
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        # license
        'License :: OSI Approved :: BSD License',
        # programming language
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        # topic
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Scientific/Engineering',
        'Topic :: Text Processing',
    ],
)



"""Module containing the logic for describe-get-system operation"""

import sys
import re
import argparse
import operator
import shlex

from functools import partial
from textwrap import wrap

from dgspoc import version

from dgspoc.config import Data

from dgspoc.interpreter import ScriptBuilder

from dgspoc.utils import File
from dgspoc.utils import Printer
from dgspoc.utils import Text
from dgspoc.utils import DotObject
from dgspoc.utils import parse_template_result
from dgspoc.utils import Misc
from dgspoc.utils import MiscArgs

from dgspoc.constant import ECODE
from dgspoc.constant import CONVTYPE
from dgspoc.constant import COMMAND
from dgspoc.constant import FEATURE

from dgspoc.storage import TemplateStorage

from dgspoc.usage import validate_usage
from dgspoc.usage import validate_example_usage
from dgspoc.usage import show_usage
from dgspoc.usage import get_global_usage

from dgspoc.adaptor import Adaptor

from dgspoc.parser import ParsedOperation
from dgspoc.parser import CheckStatement

from dgspoc.batch import BatchBuilder

from dgspoc.report import DGSReportFile
from dgspoc.report import DGSReport

from templateapp import TemplateBuilder

from dlapp import DLQuery
from dlapp import create_from_csv_data
from dlapp import create_from_json_data

from pprint import pprint
from dlapp.collection import Tabular


class OptionSelector:
    def __init__(self, options, print_help=None):
        self.options = options
        self.print_help = print_help
        self.method = None
        self.prepare()

    def prepare(self):
        if self.options.template_id.strip():
            self.method = do_clear_template
        elif self.options.filepath.strip():
            self.method = do_delete_filepath
        else:
            if self.options.command == COMMAND.USAGE:
                self.method = do_show_global_usage
            elif self.options.command == COMMAND.INFO:
                self.method = do_show_info
            elif self.options.command == COMMAND.BUILD:
                feature = ''.join(self.options.operands[:1]).lower()
                if feature == FEATURE.TEMPLATE:
                    self.method = do_build_template
                elif feature == FEATURE.SCRIPT:
                    self.method = do_build_test_script
                elif feature == FEATURE.BATCH:
                    self.method = do_build_batch_script
                else:
                    exit_code = ECODE.SUCCESS if feature == 'usage' else ECODE.BAD
                    show_usage(self.options.command, exit_code=exit_code)
            elif self.options.command == COMMAND.SEARCH:
                self.method = do_search_template
            elif self.options.command == COMMAND.TEST:
                self.method = do_testing
            elif self.options.command == COMMAND.REPORT:
                self.method = do_reporting

    def process(self):
        if callable(self.method):
            self.method(self.options)
        else:
            if callable(self.print_help):
                self.print_help()
            else:
                print('*** Something is not right.  Check with developer.')
                sys.exit(ECODE.BAD)


def do_show_global_usage(options):
    if options.command == COMMAND.USAGE:
        print(get_global_usage())
        sys.exit(ECODE.SUCCESS)


def do_show_version(options):
    if options.command == 'version':
        print('{} v{}'.format(Data.console_cli_name, version))
        sys.exit(ECODE.SUCCESS)


def do_show_info(options):
    command, operands = options.command, options.operands

    name = command
    validate_usage(command, operands)
    validate_example_usage(name, operands)

    op_txt = ' '.join(operands).lower()

    lst = []
    default_lst = [
        'Describe-Get-System Proof of Concept',
        Data.get_app_info()
    ]

    is_showed_all = options.all or re.search('all', op_txt)
    is_showed_dependency = options.dependency or re.search('depend', op_txt)
    is_showed_storage = options.template_storage or re.search('template|storage', op_txt)

    if is_showed_all:
        lst.extend(default_lst)

    if is_showed_all or is_showed_dependency:
        lst and lst.append('--------------------')
        lst.append('Packages:')
        values = Data.get_dependency().values()
        for pkg in sorted(values, key=lambda item: item.get('package')):
            lst.append('  + Package: {0[package]}'.format(pkg))
            lst.append('             {0[url]}'.format(pkg))

    if is_showed_all or is_showed_storage:
        lst and lst.append('--------------------', )
        lst.append(Data.get_template_storage_info())

    Printer.print(lst) if lst else Printer.print(default_lst)
    sys.exit(ECODE.SUCCESS)


def do_build_template(options):
    command, operands = options.command, list(options.operands)
    op_count = len(operands)
    feature = str(operands[0]).lower().strip() if op_count > 0 else ''

    operands = operands[1:]
    name = '{}_{}'.format(command, feature)
    validate_usage(name, operands)
    validate_example_usage(name, operands)

    op_txt = ' '.join(operands).rstrip()

    if not op_txt:
        show_usage(name, exit_code=ECODE.BAD)

    if File.is_exist(op_txt):
        with open(op_txt) as stream:
            user_data = stream.read()
    else:
        user_data = op_txt

    try:
        factory = TemplateBuilder(
            user_data=user_data, author=options.author, email=options.email,
            company=options.company
        )

        template_id = options.tmplid.strip()
        filename = options.filename.strip()

        fmt1 = '+++ Successfully uploaded generated template to "{}" template ID.'
        fmt2 = '+++ Successfully saved generated template to {}'
        fmt3 = 'CANT save generated template to existing {} file.  Use replaced flag accordingly.'

        if template_id or filename:
            is_ok = True
            lst = []
            if template_id:
                is_uploaded = TemplateStorage.upload(
                    template_id, factory.template, replaced=options.replaced
                )
                is_ok &= is_uploaded
                msg = fmt1.format(template_id) if is_uploaded else TemplateStorage.message
                lst.append(msg)
            if filename:
                filename = File.get_path(filename)
                if File.is_exist(filename) and not options.replaced:
                    msg = fmt3.format(filename)
                    is_ok &= False
                else:
                    is_saved = File.save(options.filename, factory.template)
                    is_ok &= is_saved
                    msg = fmt2.format(filename) if is_saved else File.message

                lst and lst.append('=' * 20)
                lst.append(msg)

            lst and Printer.print(lst)
            sys.exit(ECODE.SUCCESS if is_ok else ECODE.BAD)
        else:
            print(factory.template)
            sys.exit(ECODE.SUCCESS)

    except Exception as ex:
        not options.quiet and print(Text(ex))
        sys.exit(ECODE.BAD)


def do_search_template(options):
    command, operands = options.command, list(options.operands)
    op_count = len(operands)
    feature = str(operands[0]).lower().strip() if op_count > 0 else ''
    if command == 'search' and feature == 'template':
        operands = operands[1:]
        name = '{}_{}'.format(command, feature)
        validate_usage(name, operands)
        validate_example_usage(name, operands)

        op_txt = ' '.join(operands).rstrip()

        if not op_txt:
            show_usage(name, exit_code=ECODE.BAD)

        tmpl_id_pattern = operands[0]
        is_found = TemplateStorage.search(tmpl_id_pattern)
        print(TemplateStorage.message)
        sys.exit(ECODE.SUCCESS if is_found else ECODE.BAD)

    elif command == 'search' and feature != 'template':
        exit_code = ECODE.SUCCESS if feature == 'usage' else ECODE.BAD
        show_usage('{}_template'.format(command), exit_code=exit_code)


def do_clear_template(options):
    template_id = options.template_id.strip()

    if re.match(r'(?i) *([*]|_+all_+) *$', template_id):
        fmt = ('*** Failed to clear %r template reference because '
               'Describe-Get-System prohibits multiple clearing')
        not options.quiet and print(fmt % template_id)
        sys.exit(ECODE.BAD)

    if template_id:
        result = MiscArgs.get_parsed_result_as_data_or_file(data=template_id)
        if result.is_file:
            is_deleted = File.delete(result.filename)
            not options.quiet and print('%s %s' % ('+++' if is_deleted else '***', File.message))
            sys.exit(ECODE.SUCCESS if is_deleted else ECODE.BAD)
        else:
            is_cleared = TemplateStorage.clear(template_id)
            if is_cleared:
                fmt = '+++ Successfully cleared %r template id in template storage.'
                message = fmt % template_id
            else:
                message = '*** %s' % TemplateStorage.message
            not options.quiet and print(message)
            sys.exit(ECODE.SUCCESS if is_cleared else ECODE.BAD)


def validate_test_data_flag(options):
    command = options.command
    feature = options.operands[0] if options.operands else ''
    if options.testfile == '' and options.adaptor == '':
        lst = ['CANT run {} test WITHOUT test data.'.format(feature),
               'Please use --test-file=<test-file-name> or',
               '           --adaptor=<adaptor_name> --execution="<device cmdline>"']
        Printer.print(lst)
        show_usage('{}_{}'.format(command, feature), exit_code=ECODE.BAD)


def get_test_from_adaptor(options):
    command = options.command
    feature = options.operands[0] if options.operands else ''
    name = '{}_{}'.format(command, feature)
    execution = options.execution.strip()
    if not execution:
        lst = [
            'ExecutionSyntaxError: must be',
            '--execution="--host=<addr_or_name> <cmdline>"',
        ]
        Printer.print(lst)
        show_usage(name, exit_code=ECODE.BAD)

    try:

        lst = execution.split(' ')

        parser = argparse.ArgumentParser(exit_on_error=False)
        parser.add_argument('items', nargs='*')
        parser.add_argument('--host', type=str, default='')
        parser.add_argument('other_items', nargs='*')
        parser_args = parser.parse_args(lst)

        host = parser_args.host
        cmdline = ' '.join(parser_args.items + parser_args.other_items)

        if not parser_args.host:
            lst = [
                'ExecutionSyntaxError: must be',
                '--execution="--host=<addr_or_name> <cmdline>"',
            ]
            Printer.print(lst)
            show_usage(name, exit_code=ECODE.BAD)

        device = Adaptor(options.adaptor, host)
        device.connect()
        test_data = device.execute(cmdline)
        device.disconnect()
        device.release()
        return test_data
    except Exception as ex:
        failure = 'AdaptorInquiryError - ({})'.format(Text(ex))
        not options.quiet and Printer.print(failure)
        sys.exit(ECODE.BAD)


def do_testing(options):
    pattern = '^--(template-id|(file(-?name)?))='
    command, operands = options.command, list(options.operands)
    adaptor = options.adaptor.strip().lower()
    action = options.action.strip()

    name = command
    validate_usage(name, operands)
    validate_example_usage(name, operands)
    is_showed = False
    if re.search(r'(?i)--showed\b', action):
        is_showed = True
        action = re.sub(r'(?i)--showed\b', '', action).strip()

    is_tabular = False
    if re.search(r'(?i)--tabular\b', action):
        is_tabular = True
        action = re.sub(r'(?i)--tabular\b', '', action).strip()

    adaptor = adaptor or 'stream'
    is_adaptor_stream = bool(re.match(r'(stream|file(name)?)$', adaptor))
    is_execute_cmdline = CheckStatement.is_execute_cmdline(action)

    if is_adaptor_stream:
        data = 'dummy execute %s' % action if not is_execute_cmdline else action

        node = ParsedOperation(data)
        test_data_file = re.sub(pattern, '', node.operation_ref.strip())
        File.message = ''
        output = File.get_content(test_data_file)
        if File.message:
            not options.quiet and print('*** %s' % File.message)
            sys.exit(ECODE.BAD)
        print(output)
        result = DotObject(test_data=output, template='', records=[], records_count=0)

        if node.has_select_statement or node.convertor:
            if node.is_csv or node.is_json:
                try:
                    lines = result.test_data.splitlines()
                    index = 0
                    pat = (r'(?i)\w{3} +\d\d? +\d{4} '
                           r'\d\d:\d\d:\d\d[.]\d\d\d for "\S+" - '
                           r'UNREAL-DEVICE-\w+-SERVICE-TIMESTAMP')
                    for i, line in enumerate(lines):
                        if re.match(pat, line):
                            index = i
                            break
                    test_data = '\n'.join(lines[index+1:])
                    method = create_from_json_data if node.is_json else create_from_csv_data
                    records = method(test_data).data
                    result.records = records
                    result.records_count = len(records)
                except Exception as ex:
                    not options.quiet and print('*** %s' % Text(ex))
                    sys.exit(ECODE.BAD)
            else:
                pfunc = partial(parse_template_result, test_file=test_data_file)
                if not node.convertor == CONVTYPE.TEMPLATE or not node.convertor_arg:
                    print('*** Invalid action: %s ***' % action)
                    sys.exit(ECODE.BAD)

                try:
                    if re.match('(?i)--file', node.convertor_arg.strip()):
                        template_file = re.sub(pattern, '', node.convertor_arg.strip())
                        result = pfunc(template_file=template_file)
                    else:
                        tmpl_id = re.sub(pattern, '', node.convertor_arg.strip())
                        if TemplateStorage.check(tmpl_id):
                            tmpl_data = TemplateStorage.get(tmpl_id)
                            result = pfunc(template_data=tmpl_data)
                        else:
                            fmt = '*** %r template id CANT find in template storage.'
                            print(fmt % tmpl_id)
                            sys.exit(ECODE.BAD)
                except Exception as ex:
                    failure = '*** %s' % ex
                    not options.quiet and print(failure)
                    sys.exit(ECODE.BAD)
        else:
            sys.exit(ECODE.SUCCESS)
    else:
        if not CheckStatement.is_performer_statement(action):
            failure = '*** Invalid action: %s ***' % action
            print(failure)
            sys.exit(ECODE.BAD)

        node = ParsedOperation(action)
        if not node.devices_names:
            failure = '*** Invalid action: %s ***' % action
            print(failure)
            sys.exit(ECODE.BAD)

        host = node.devices_names[0]
        connection = Adaptor(adaptor, host)
        tbl = dict(execution=connection.execute,
                   configuration=connection.configure,
                   reload=connection.reload)
        operation_method = tbl[node.name]
        connection.connect()
        output = operation_method(node.operation_ref)
        connection.release()

        if node.is_reload or node.is_configuration:
            print(output)
            sys.exit(ECODE.SUCCESS)

        if node.is_csv or node.is_json:
            try:
                lines = output.splitlines()
                index = 0
                pat = (r'(?i)\w{3} +\d\d? +\d{4} '
                       r'\d\d:\d\d:\d\d[.]\d\d\d for "\S+" - '
                       r'UNREAL-DEVICE-\w+-SERVICE-TIMESTAMP')
                for i, line in enumerate(lines):
                    if re.match(pat, line):
                        index = i
                        break
                test_data = '\n'.join(lines[index + 1:])
                method = create_from_json_data if node.is_json else create_from_csv_data
                records = method(test_data).data
                result = DotObject(
                    test_data=output, template='',
                    records=records, records_count=len(records)
                )
            except Exception as ex:
                not options.quiet and print('*** %s' % Text(ex))
                sys.exit(ECODE.BAD)
        elif node.is_template:
            pfunc = partial(parse_template_result, test_data=output)
            try:
                if re.match('(?i)--file', node.convertor_arg.strip()):
                    template_file = re.sub(pattern, '', node.convertor_arg.strip())
                    result = pfunc(template_file=template_file)
                else:
                    tmpl_id = re.sub(pattern, '', node.convertor_arg.strip())
                    if TemplateStorage.check(tmpl_id):
                        tmpl_data = TemplateStorage.get(tmpl_id)
                        result = pfunc(template_data=tmpl_data)
                    else:
                        fmt = '*** %r template id CANT find in template storage.'
                        print(fmt % tmpl_id)
                        sys.exit(ECODE.BAD)
            except Exception as ex:
                failure = '*** %s' % ex
                not options.quiet and print(failure)
                sys.exit(ECODE.BAD)
        else:
            if not node.has_select_statement:
                sys.exit(ECODE.SUCCESS)
            else:
                failure = '*** Invalid action: %s ***' % action
                print(failure)
                sys.exit(ECODE.BAD)

    if node.has_select_statement:
        query_obj = DLQuery(result.records)
        tested_records = query_obj.find(select=node.select_statement)
    else:
        tested_records = result.records

    if is_showed:
        Printer.print('User Test Data:')
        print(result.test_data)
        print()

        if result.template:
            Printer.print('Template:')
            print('%s\n' % result.template)
            print()

    lst = ['Parsed Results:']
    if node.has_select_statement:
        lst.append('    SELECT-STATEMENT: %s' % node.select_statement)
    Printer.print(lst)
    if is_tabular:
        Tabular(tested_records).print()
    else:
        pprint(tested_records)
    print()

    if node.is_need_verification and Misc.is_list(tested_records):
        total = len(tested_records)
        op = node.condition
        expected_number = node.expected_condition
        chk = getattr(operator, op)(total, expected_number)
        prefix = '' if chk else '*** CANT BE ***'
        tbl = dict(eq='==', ne='!=', lt='<', le='<=', gt='>', ge='>=')

        lst = ['Verification:',
               '    CONDITION: %s' % node.condition_data,
               '    STATUS   : %s' % ('Passed' if chk else 'Failed')]
        Printer.print(lst)

        fmt = '%s (total found records: %s) %s (expected total count: %s)'
        msg = (fmt % (prefix, total, tbl[op], expected_number)).strip()
        print(msg)
        print()

    if node.error:
        lst = ['Error:', '------', '\n'.join(wrap(node.error, width=76)), '']
        not options.quiet and Printer.print(lst)
        sys.exit(ECODE.BAD)
    else:
        sys.exit(ECODE.SUCCESS)


def do_build_test_script(options):
    command, operands = options.command, list(options.operands)

    feature = ''.join(operands[:1])
    operands = operands[1:]
    framework = 'unittest'
    for node in [FEATURE.UNITTEST, FEATURE.PYTEST, FEATURE.ROBOTFRAMEWORK]:
        if node == feature:
            framework = str(node)
            break

    name = '{}_script'.format(command)
    validate_usage(name, operands)
    validate_example_usage(name, operands)

    if len(operands) > 1:
        Printer.print('*** Application only support single script snippet.')
        show_usage(name, exit_code=ECODE.BAD)

    snippet_filename = operands[0]
    snippet_content = File.get_content(snippet_filename)
    if File.message:
        not options.quiet and Printer.print('*** %s' % File.message)
        sys.exit(ECODE.BAD)

    node = ScriptBuilder(
        snippet_content, framework=framework, is_logger=True,
        username=options.author, email=options.email, company=options.company
    )

    test_script = node.testscript.strip()

    if not test_script:
        fmt = '*** No script is generated for %r'
        Printer.print(fmt % snippet_filename)
        sys.exit(ECODE.BAD)

    if options.filename:
        File.save(options.filename, test_script)
        if File.message and not File.message.startswith('Successfully saved data'):
            fmt = '*** Failed to save the generated test script to %s\n*** %s'
            Printer.print(fmt % (options.filename, File.message))
            sys.exit(ECODE.BAD)
        else:
            fmt = '+++ Successfully saved the generated test script to %s'
            Printer.print(fmt % options.filename)
            sys.exit(ECODE.SUCCESS)
    else:
        print(test_script)
        sys.exit(ECODE.SUCCESS)


def do_build_batch_script(options):
    command, operands = options.command, list(options.operands)

    # feature = ''.join(operands[:1])
    operands = operands[1:]
    name = '{}_batch'.format(command)
    validate_usage(name, operands)
    validate_example_usage(name, operands)

    builder = BatchBuilder(options)
    batch_script = builder.script
    if batch_script:
        if options.filename:
            filename = options.filename

            batch_script += '\ndgs report --detail' if options.detail else '\ndgs report'
            batch_script += '\ndgs --delete=dgs_test_script_files --quiet'

            is_saved = File.save(filename, batch_script)
            if is_saved:
                fmt = '+++ Successfully saved %r batch file.'
                not options.quiet and print(fmt % filename)
                sys.exit(ECODE.SUCCESS)
            else:
                fmt = '*** Failed to save %r batch file.'
                not options.quiet and print(fmt % filename)
                not options.quiet and print('*** %s' % File.message)
                sys.exit(ECODE.BAD)
        else:
            print(batch_script)
            sys.exit(ECODE.SUCCESS)
    else:
        msg = '*** Failed to generate batch content because no test file is found.'
        not options.quiet and print(msg)
        sys.exit(ECODE.BAD)


def do_delete_filepath(options):
    filepath = options.filepath

    if File.is_exist(filepath):
        is_deleted = File.delete(filepath)
        fmt = '+++ %s' if is_deleted else '*** %s'
        not options.quiet and print(fmt % File.message)
        sys.exit(ECODE.SUCCESS if is_deleted else ECODE.BAD)

    are_deleted = None
    posix = False if Misc.is_window_os() else True
    for file_path in shlex.split(filepath, posix=posix):
        if file_path == ',':
            continue

        file_path = file_path.rstrip(',').strip('{').strip('}')
        is_deleted = File.delete(file_path)
        fmt = '+++ %s' if is_deleted else '*** %s'
        not options.quiet and print(fmt % File.message)
        are_deleted = is_deleted if are_deleted is None else are_deleted and is_deleted

    if are_deleted is None:
        sys.exit(ECODE.BAD)
    else:
        sys.exit(ECODE.SUCCESS if are_deleted else ECODE.BAD)


def do_reporting(options):
    command, operands = options.command, list(options.operands)
    # feature = ''.join(operands[:1])
    # operands = operands[1:]
    name = command
    validate_usage(name, operands)
    validate_example_usage(name, operands)

    directory = operands[0] if operands else '.'
    report_files = DGSReportFile.get_report_files(directory)
    reporter = DGSReport(*report_files, detail=options.detail)
    report = reporter.generate()
    print(report)
    sys.exit(reporter.exit_code)
